from __future__ import annotations

import datetime
import enum
import json
import logging
import os
import pathlib
import re
import tomllib
import typing as _t

import packaging.version
import semver
import yuio.app
import yuio.complete
import yuio.git
import yuio.io
import yuio.parse
import yuio.theme
from markdown_it.tree import SyntaxTreeNode

from cl_keeper._version import __version__
from cl_keeper.check import check as _check
from cl_keeper.config import (
    PYTHON_PRESET,
    Config,
    GlobalConfig,
    LinkTemplates,
    VersionFormat,
)
from cl_keeper.context import Context, IssueCode, IssueScope
from cl_keeper.fix import fix as _fix
from cl_keeper.model import (
    Changelog,
    ReleaseSection,
    RepoVersion,
    Section,
    SubSection,
    SubSectionCategoryKind,
    UnreleasedSection,
    Version,
)
from cl_keeper.parse import canonize_version as _canonize_version
from cl_keeper.parse import detect_subsection_metadata as _detect_subsection_metadata
from cl_keeper.parse import parse as _parse
from cl_keeper.parse import parse_version as _parse_version
from cl_keeper.parse import split_into_sections as _split_into_sections
from cl_keeper.render import print_diff as _print_diff
from cl_keeper.render import render as _render
from cl_keeper.sort import merge_sections as _merge_sections
from cl_keeper.vcs import detect_origin as _detect_origin
from cl_keeper.vcs import get_repo_versions

logger = logging.getLogger(__name__)


_GLOBAL_OPTIONS: GlobalConfig = GlobalConfig()

_TRAILER_RE = re.compile(
    r"^\s*(?:\[(?P<group>[^]]*+)\])?\s*(?P<message>.*)$", re.MULTILINE | re.DOTALL
)
_COMMENT_RE = re.compile(r"\<\!\-\-.*?(\-\-\>|\Z)", re.MULTILINE | re.DOTALL)


@yuio.app.app(version=__version__)
def main(
    #: override path to the config file
    config_path: (
        _t.Annotated[
            pathlib.Path,
            yuio.parse.ExistingPath(extensions=[".toml", ".yaml", ".yml"]),
        ]
        | None
    ) = yuio.app.field(default=None, flags=["-c", "--config"], usage=yuio.OMIT),
    #: override path to the changelog file
    file: pathlib.Path | None = yuio.app.field(
        default=None,
        flags=["-i", "--input"],
        completer=yuio.complete.File(extensions=[".md"]),
        usage=yuio.OMIT,
    ),
    #: increase severity of all messages by one level
    strict: bool = yuio.app.field(default=False, usage=yuio.OMIT),
    #: config overrides
    cfg: Config = yuio.app.field(usage=yuio.OMIT),
):
    _GLOBAL_OPTIONS.config_path = config_path
    _GLOBAL_OPTIONS.strict = strict
    _GLOBAL_OPTIONS.file = file

    logging.getLogger("markdown_it").setLevel("WARNING")

    logger.debug("changelog keeper version %s", __version__)


main.description = """
a helper for maintaining changelog files that use `keep-a-changelog` format.

"""

main.epilog = """
# further help:

- to get help for a specific subcommand:

  ```sh
  chk <subcommand> --help
  ```

- online documentation: https://cl-keeper.readthedocs.io/.

- changelog format: https://keepachangelog.com/
"""


@main.subcommand
def check():
    """
    check contents of the changelog file.

    """

    config = _load_config()
    file = _locate_changelog(config)

    ctx = Context(
        file,
        file.read_text(),
        config,
        _GLOBAL_OPTIONS.strict,
        _make_link_templates(file.parent, config),
    )

    repo_versions = None
    if ctx.config.check_repo_tags:
        repo_versions = get_repo_versions(file.parent, ctx)

    _check(_parse(ctx), ctx, repo_versions)

    if ctx.has_messages():
        ctx.report()
        ctx.exit_if_has_errors()
    else:
        yuio.io.success("No issues detected")


@main.subcommand
def fix(
    #: don't save changes, print the diff instead
    dry_run: bool = False,
    #: print diff
    diff: bool = False,
):
    """
    fix contents of the changelog file.

    """

    config = _load_config()
    file = _locate_changelog(config)
    original = file.read_text()

    ctx = Context(
        file,
        original,
        config,
        _GLOBAL_OPTIONS.strict,
        _make_link_templates(file.parent, config),
    )

    repo_versions = None
    if ctx.config.check_repo_tags:
        repo_versions = get_repo_versions(file.parent, ctx)

    changelog = _parse(ctx)
    _fix(changelog, ctx, repo_versions)

    result = _render(changelog, ctx)

    ctx.reset(result)

    _check(_parse(ctx), ctx, repo_versions)

    if result == original and ctx.has_messages():
        yuio.io.success("No fixable issues detected")
    elif result == original:
        yuio.io.success("No issues detected")
    elif not dry_run:
        yuio.io.success("Changelog successfully fixed")
        file.write_text(result)

    if ctx.has_messages():
        yuio.io.heading("Unfixed issues")
        ctx.report()

    if dry_run or diff:
        _print_diff(original, result, file)


class BumpMode(enum.Enum):
    AUTO = "auto"
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"
    POST = "post"

    def __str__(self) -> str:
        return self.value


_PRE_RELEASE_GROUP = yuio.app.MutuallyExclusiveGroup()


@main.subcommand
def bump(
    #: produce result even if errors are detected
    ignore_errors: bool = False,
    #: don't save changes, print the diff instead
    dry_run: bool = False,
    #: release version or tag that will be used for a new release
    version: (
        _t.Annotated[BumpMode, yuio.parse.WithMeta(desc="version component")]
        | str
        | None
        | yuio.git.Tag
    ) = yuio.app.positional(default=None),
    #: create an alpha pre-release. If `version` is given, this will bump
    #: the corresponding component and make a pre-release. If `version` is not given,
    #: the latest release must be a pre-release itself.
    alpha: bool = yuio.app.field(default=False, group=_PRE_RELEASE_GROUP),
    #: create a beta pre-release, similar to `--alpha`.
    beta: bool = yuio.app.field(default=False, group=_PRE_RELEASE_GROUP),
    #: create a release candidate, similar to `--alpha`.
    rc: bool = yuio.app.field(default=False, group=_PRE_RELEASE_GROUP),
    #: open generated changes for editing
    edit: bool = False,
    #: commit and tag the release after updating changelog.
    commit: bool = False,
):
    """
    move entries from `unreleased` to a new release.

    """

    if version is None and not alpha and not beta and not rc:
        raise yuio.parse.ParsingError(
            "<version> is required when --alpha, --beta, and --rc are not specified"
        )
    if isinstance(version, str) and (alpha or beta or rc):
        raise yuio.parse.ParsingError(
            "--alpha, --beta, --rc are not allowed when specifying custom versions"
        )

    config = _load_config()
    file = _locate_changelog(config)
    original = file.read_text()

    if commit:
        repo = yuio.git.Repo(file.parent)
        status = repo.status()
        if status.cherry_pick_head:
            raise yuio.app.AppError("Can't bump changelog: cherry pick is in progress")
        if status.merge_head:
            raise yuio.app.AppError("Can't bump changelog: merge is in progress")
        if status.rebase_head:
            raise yuio.app.AppError("Can't bump changelog: rebase is in progress")
        if status.revert_head:
            raise yuio.app.AppError("Can't bump changelog: revert is in progress")
        if not status.branch:
            raise yuio.app.AppError("Can't bump changelog: git head is detached")

    ctx = Context(
        file,
        original,
        config,
        _GLOBAL_OPTIONS.strict,
        _make_link_templates(file.parent, config),
    )

    if isinstance(version, str) and version.startswith(config.tag_prefix):
        version = version[len(config.tag_prefix) :]

    changelog = _parse(ctx)

    found = None
    to_remove = set()
    for i, section, _ in _find_sections(FindMode.UNRELEASED, changelog, config):
        to_remove.add(i)
        if found is None:
            found = section
        else:
            _merge_sections(found, section)

    changelog.sections = [
        section for i, section in enumerate(changelog.sections) if i not in to_remove
    ]

    if found is None:
        found = UnreleasedSection(
            heading=None, subsections=[], version_link=None, version_label=None
        )

    repo_versions = None
    if isinstance(version, BumpMode) or version is None:
        repo_versions = get_repo_versions(file.parent, ctx)
        latest_version, _, _ = _find_latest_version(changelog, repo_versions, config)
        if latest_version is None:
            raise yuio.app.AppError("No previous release to bump")
        version = _bump_version(latest_version, version, alpha, beta, rc, found, config)
    else:
        if ctx.config.check_repo_tags:
            repo_versions = get_repo_versions(file.parent, ctx)

    parsed_version = _parse_version(version, ctx.config)
    if parsed_version is None and config.version_format is not VersionFormat.NONE:
        ctx.issue(
            IssueCode.INVALID_VERSION,
            "New version `%s` doesn't follow %s specification",
            version,
            ctx.config.version_format.value,
            scope=IssueScope.EXTERNAL,
        )

    _check(changelog, ctx, repo_versions)

    if ctx.has_errors():
        ctx.report()
    if not ignore_errors:
        ctx.exit_if_has_errors()

    for _, section, _ in _find_sections(version, changelog, config):
        if section.map is not None:
            pos = f" on line `{section.map[0] + 1}`"
        else:
            pos = ""
        raise yuio.app.AppError(f"release {version} already exists{pos}")

    new_section = ReleaseSection(
        heading=None,
        subsections=found.subsections,
        version=version,
        parsed_version=parsed_version,
        canonized_version=_canonize_version(parsed_version, config) or version,
        release_date=datetime.date.today(),
        version_link=None,
        version_label=None,
        release_date_fmt=datetime.date.today().isoformat(),
        release_comment=None,
    )

    if edit:
        _edit_section(changelog, ctx, new_section)

    changelog.sections.append(
        UnreleasedSection(
            heading=None, subsections=[], version_link=None, version_label=None
        )
    )
    changelog.sections.append(new_section)

    _fix(changelog, ctx, repo_versions)

    result = _render(changelog, ctx)

    if not dry_run:
        yuio.io.success("Changelog successfully updated")
        file.write_text(result)
    else:
        _print_diff(original, result, file)

    if commit:
        yuio.io.heading("Commit")
    if ctx.has_errors() and commit:
        yuio.io.warning("Not committing changes: errors detected")
        ctx.exit_if_has_errors()
    tag = f"{config.tag_prefix}{version}"
    if commit and not dry_run:
        repo = yuio.git.Repo(file.parent)
        repo.git("add", str(file))
        repo.print_status()
        yuio.io.info(
            "You can take a look around and make any changes before proceeding.\n"
            "Alternatively, you can cancel now and make a release commit later.\n"
        )
        ok = yuio.io.ask[bool]("Proceed with commit and tag?", default=False)
        message = f"Release {version}"
        if not ok:
            yuio.io.failure("Commit canceled")
            yuio.io.md(
                """
                Use this command to commit changes:

                ```sh
                git commit -m '%s'
                git tag %s
                ```
                """,
                message,
                tag,
            )
            raise yuio.app.AppError()
        repo.git("commit", "-m", message)
        repo.git("tag", tag)
        status = repo.status()
        yuio.io.success("Created commit `%s` and tag `%s`.", status.commit, tag)
        yuio.io.md(
            """
            Use `git commit --amend` and `git tag` to update commit contents or message:

            ```sh
            git commit --amend && git tag -f %s
            ```
            """,
            tag,
        )
    if commit and dry_run:
        yuio.io.success("Dry-run: not committing and creating tag `%s`", tag)


bump.usage = """
%(prog)s [<options>] [--ignore-errors] [--dry-run] [--edit] <version>
%(prog)s [<options>] [--ignore-errors] [--dry-run] [--edit] [auto|major|minor|patch|post] [--alpha|--beta|--rc]
"""

bump.epilog = """
# examples:

## specifying version manually

you can always supply a specific version for a new release. As long as there are no
prior releases with the same version, this operation will succeed:

```sh
chk bump 1.0.5
```

## automatic version detection

changelog keeper can suggest next version based on changelog content
(the `--dry-run` flag is handy here):

```sh
chk bump auto
```

## bumping a version component

you can bump a version component by setting `<version>` to `major`, `minor`,
`patch`, or `post`:

```sh
chk bump minor  # 1.0.0 -> 1.1.0
```

## pre-releases

you can make a pre-release by adding `--alpha`, `--beta`, or `--rc`:

```sh
chk bump major --beta  # 1.5.1 -> 2.0.0b0
```

if the last release is a pre-release itself, you can bump its pre-release version
component:

```sh
chk bump --beta  # 1.0.0b0 -> 1.0.0b1
chk bump --rc  # 1.0.0b0 -> 1.0.0rc0
```

"""


class FindMode(enum.Enum):
    LATEST = "latest"
    UNRELEASED = "unreleased"

    def __str__(self) -> str:
        return self.value


@main.subcommand
def find(
    #: produce result even if errors are detected
    ignore_errors: bool = False,
    #: print data in JSON format
    format_json: bool = yuio.app.field(default=False, flags="--json"),
    #: release version or tag, can also be `unreleased` or `latest`
    version: (
        _t.Annotated[FindMode, yuio.parse.WithMeta(desc="query")] | str | yuio.git.Tag
    ) = yuio.app.positional(),
):
    """
    find a changelog entry for a given release version.

    """

    config = _load_config()
    file = _locate_changelog(config)
    original = file.read_text()

    if isinstance(version, str):
        version = version.removeprefix("refs/tags/").removeprefix(config.tag_prefix)

    ctx = Context(
        file,
        original,
        config,
        _GLOBAL_OPTIONS.strict,
        _make_link_templates(file.parent, config),
    )

    repo_versions = None
    if ctx.config.check_repo_tags:
        repo_versions = get_repo_versions(file.parent, ctx)

    changelog = _parse(ctx)

    _check(changelog, ctx, repo_versions)

    if ctx.has_errors():
        ctx.report()
    if not ignore_errors:
        ctx.exit_if_has_errors()

    found = None
    found_latest_in_changelog = None
    for _, section, is_latest in _find_sections(version, changelog, config):
        if found is None:
            found = section
            found_latest_in_changelog = is_latest
        else:
            _merge_sections(found, section)

    if found is None and isinstance(version, str):
        parsed_version = _parse_version(version, config)
        canonized_version = _canonize_version(parsed_version, config) or version

        lower_bound = ctx.config.parsed_ignore_missing_releases_before
        regex_bound = ctx.config.ignore_missing_releases_regexp

        if (repo_versions is None or canonized_version in repo_versions) and (
            (lower_bound and parsed_version and parsed_version < lower_bound)
            or (regex_bound is not None and re.search(regex_bound, version))
        ):
            found = ReleaseSection(
                heading=None,
                subsections=[],
                version=version,
                parsed_version=parsed_version,
                canonized_version=canonized_version,
                version_link=None,
                version_label=None,
                release_date=None,
                release_date_fmt=None,
                release_comment=None,
            )

    if found is not None:
        tokens = found.to_tokens(include_heading=False)
        if format_json:
            found_version = None
            found_canonized_version = None
            found_latest = None
            tag = None
            is_pre_release = None
            is_post_release = None
            if release := found.as_release():
                found_version = release.version
                found_canonized_version = release.canonized_version
                if repo_versions and (
                    repo_version := repo_versions.get(release.canonized_version)
                ):
                    tag = f"{config.tag_prefix}{repo_version.version}"
                is_pre_release = _is_pre_release(release.parsed_version)
                is_post_release = _is_post_release(release.parsed_version)
                try:
                    _, _, latest_version_canonized = _find_latest_version(
                        changelog, repo_versions, config
                    )
                except:
                    pass
                else:
                    found_latest = release.canonized_version == latest_version_canonized
            data = {
                "version": found_version,
                "canonizedVersion": found_canonized_version,
                "tag": tag,
                "text": _render(changelog, ctx, tokens, disable_wrapping=True),
                "isLatestInChangelog": found_latest_in_changelog,
                "isLatestInSemanticOrder": found_latest,
                "isPreRelease": is_pre_release,
                "isPostRelease": is_post_release,
                "isUnreleased": found.is_unreleased(),
            }
            print(json.dumps(data, indent="  "))
        else:
            print(_render(changelog, ctx, tokens, disable_wrapping=True), end="")
    else:
        raise yuio.app.AppError("Can't find changelog entry for version `%s`", version)


find.usage = """
%(prog)s [<options>] [--ignore-errors] [--json] <version>
%(prog)s [<options>] [--ignore-errors] [--json] {latest|unreleased}
"""


find.epilog = """
# examples:

find an exact version:

```sh
chk find v1.0.0-rc2
```

find the first version (barring the `unreleased` section)
that appears in the changelog:

```sh
chk find latest
```

find the `unreleased` section:

```sh
chk find unreleased
```

# json output:

If `--json` flag is given, a JSON object is printed to stdout. It will contain
the following fields:

- <c hl/flag:sh-usage>version</c> - version string, as appears in the changelog.

  Can be `null` if unreleased section is requested.

- <c hl/flag:sh-usage>canonizedVersion</c> - version string, canonized according to the used
  `version_format`.

  If `version_format` is `null` or canonization fails, this will contain
  string from `version`.

- <c hl/flag:sh-usage>tag</c> - tag that corresponds to this version.

  Can be `null` if requested release not found, if unreleased section is requested,
  if `check_repo_tags` is `false`, if version canonization fails, or if tag is not
  found for this release.

- <c hl/flag:sh-usage>text</c> - text extracted from the changelog entry.

  Can be empty if requested release not found.

- <c hl/flag:sh-usage>isLatestInChangelog</c> - `true` if found version appears first in the changelog file.

  Can be `null` if requested release not found, or if unreleased section
  is requested.

- <c hl/flag:sh-usage>isLatestInSemanticOrder</c> - `true` if this is the latest known release.

  Release versions are compared with respect to the selected `version_format`.
  If `check_repo_tags` is `true`, this flag also checks all tags found
  in the repository.

  Can be `null` if requested release not found, or if unreleased section
  is requested.

- <c hl/flag:sh-usage>isPreRelease</c> - `true` if release version contains a pre-release component,
  like `beta` or `rc`.

  Can be `null` if requested release not found, if unreleased section
  is requested, or if version canonization fails.

- <c hl/flag:sh-usage>isPostRelease</c> - `true` if release version contains a post-release component.

  Can be `null` if requested release not found, if unreleased section
  is requested, or if version canonization fails.

- <c hl/flag:sh-usage>isUnreleased</c> - `true` if unreleased section is requested.

"""


@main.subcommand
def check_tag(
    #: full name of the tag to check
    tag: yuio.git.Tag | str = yuio.app.positional(),
):
    """
    check if a git tag conforms to the versioning specification.

    this command is handy to use in release CI or in a pre-push git hook to verify
    that all tags conform to the selected versioning specification.

    """

    config = _load_config()
    error = False
    if not tag.startswith(config.tag_prefix):
        yuio.io.error("Tag `%s` should start with `%r`", tag, config.tag_prefix)
        error = True
    parsed = _parse_version(tag.removeprefix(config.tag_prefix), config)
    if not parsed:
        yuio.io.error(
            "Tag `%s` does not follow %s specification",
            tag,
            config.version_format.value,
        )
        error = True
    if error:
        raise yuio.app.AppError("Tag verification failed")
    else:
        yuio.io.success("Tag `%s` is valid", tag)


@main.subcommand(help=yuio.DISABLED)
def pre_commit_check(
    diff: bool = False, changed_files: set[pathlib.Path] = yuio.app.positional()
):
    """
    same as `fix`, but expects list of changed files as command line arguments,
    and doesn't perform fix if changelog is not updated.

    """
    config = _load_config()
    file = _locate_changelog(config)

    if file not in changed_files and (
        _GLOBAL_OPTIONS.config_path is None
        or _GLOBAL_OPTIONS.config_path not in changed_files
    ):
        return

    fix.command(diff=diff)


@main.subcommand(help=yuio.DISABLED)
def pre_commit_check_tag():
    """
    same as `check-tag`, but tag is supplied via `PRE_COMMIT_REMOTE_BRANCH` env var.

    """
    remote_branch = os.environ.get("PRE_COMMIT_REMOTE_BRANCH")
    if not remote_branch:
        return
    config = _load_config()
    prefix = f"refs/tags/{config.tag_prefix}"
    if not remote_branch.startswith(prefix):
        return
    check_tag.command(remote_branch.removeprefix("refs/tags/"))


def _load_config() -> Config:
    config = Config()
    if _GLOBAL_OPTIONS.config_path is None:
        root = _GLOBAL_OPTIONS.file or pathlib.Path.cwd()
        while root:
            changelog_toml = root.joinpath(".changelog.toml")
            changelog_toml_exists = changelog_toml.exists()
            changelog_yaml = root.joinpath(".changelog.yaml")
            changelog_yaml_exists = changelog_yaml.exists()
            changelog_yml = root.joinpath(".changelog.yml")
            changelog_yml_exists = changelog_yml.exists()
            pyproject_toml = root.joinpath("pyproject.toml")
            pyproject_toml_exists = pyproject_toml.exists()

            if changelog_toml_exists + changelog_yaml_exists + changelog_yml_exists > 1:
                found = ", ".join(
                    f"<c path>{file}</c>"
                    for file, exists in [
                        (changelog_toml, changelog_toml_exists),
                        (changelog_yaml, changelog_yaml_exists),
                        (changelog_yml, changelog_yml_exists),
                    ]
                    if exists
                )
                raise yuio.app.AppError(f"Found multiple config files: {found}")
            if changelog_toml_exists:
                _GLOBAL_OPTIONS.config_path = changelog_toml
                break
            elif changelog_yaml_exists:
                _GLOBAL_OPTIONS.config_path = changelog_yaml
                break
            elif changelog_yml_exists:
                _GLOBAL_OPTIONS.config_path = changelog_yml
                break
            elif pyproject_toml_exists:
                _GLOBAL_OPTIONS.config_path = pyproject_toml
                break
            next_root = root.parent
            if next_root == root:
                break
            root = next_root
        if _GLOBAL_OPTIONS.config_path is None:
            logger.info("using default config")
            logger.debug("config = %r", config)
            return config
        else:
            logger.debug("found config %s", _GLOBAL_OPTIONS.config_path)
    else:
        logger.debug("loading config %s", _GLOBAL_OPTIONS.config_path)

    if not _GLOBAL_OPTIONS.config_path.exists():
        raise yuio.app.AppError(
            "Config file <c path>%s</c> doesn't exist", _GLOBAL_OPTIONS.config_path
        )
    if not _GLOBAL_OPTIONS.config_path.is_file():
        raise yuio.app.AppError(
            "Config path <c path>%s</c> is not a file", _GLOBAL_OPTIONS.config_path
        )
    if _GLOBAL_OPTIONS.config_path.name == "pyproject.toml":
        config.update(PYTHON_PRESET)
        try:
            data = tomllib.loads(_GLOBAL_OPTIONS.config_path.read_text())
        except tomllib.TOMLDecodeError as e:
            yuio.io.warning(
                "Failed to parse config file <c path>%s</c>: %s",
                _GLOBAL_OPTIONS.config_path,
                e,
            )
            logger.debug("config = %r", config)
            return config
        try:
            data = data["tool"]["cl_keeper"]
        except KeyError as e:
            logger.debug(
                "%s doesn't have section tool.cl_keeper",
                _GLOBAL_OPTIONS.config_path,
            )
            logger.info("using config from %s", _GLOBAL_OPTIONS.config_path)
            logger.debug("config = %r", config)
            return config

        config.update(
            Config.load_from_parsed_file(data, path=_GLOBAL_OPTIONS.config_path)
        )
    elif _GLOBAL_OPTIONS.config_path.suffix == ".toml":
        config.update(Config.load_from_toml_file(_GLOBAL_OPTIONS.config_path))
    elif _GLOBAL_OPTIONS.config_path.suffix in [".yaml", ".yml"]:
        config.update(Config.load_from_yaml_file(_GLOBAL_OPTIONS.config_path))
    else:
        raise yuio.app.AppError(
            "Unknown config format `%s`: <c path>%s</c>",
            _GLOBAL_OPTIONS.config_path.suffix,
            _GLOBAL_OPTIONS.config_path,
        )

    logger.info("using config from %s", _GLOBAL_OPTIONS.config_path)
    logger.debug("config = %r", config)
    return config


def _locate_changelog(config: Config) -> pathlib.Path:
    if (file := _GLOBAL_OPTIONS.file) is None:
        if _GLOBAL_OPTIONS.config_path is None:
            file = config.file
        else:
            file = _GLOBAL_OPTIONS.config_path.parent / config.file
    file = file.expanduser().resolve()
    if not file.exists():
        raise yuio.app.AppError("File `%s` doesn't exist", file)
    if not file.is_file():
        raise yuio.app.AppError("Path `%s` is not a file", file)
    return file


def _make_link_templates(repo_root: pathlib.Path, config: Config) -> LinkTemplates:
    if not config.add_release_link:
        return LinkTemplates("", "", "", {})

    link_templates = LinkTemplates(
        config.release_link_template or "",
        config.release_link_template_last or "",
        config.release_link_template_first or "",
        config.release_link_template_vars or {},
    )

    if link_templates.has_unresolved_links() and config.release_link_preset:
        link_templates.update(config.release_link_preset.get_links())

    if link_templates.has_unresolved_links():
        logger.debug("link templates are not set, trying to detect repo url from git")
        origin = _detect_origin(repo_root)
        if origin:
            link_templates.update(origin)
        else:
            raise yuio.app.AppError(
                "Can't detect url to use in link templates. Please, set "
                "`release_link_preset` or `release_link_template`, "
                "`release_link_template_last`, `release_link_template_first` "
                "in config. Alternatively, set `add_release_link` to `false`"
            )

    logger.debug("link_templates = %r", link_templates)

    _check_link_templates(link_templates)

    return link_templates


def _check_link_templates(link_templates: LinkTemplates):
    class _UnknownKeyDetector:
        def __init__(self, vars: dict[str, str]) -> None:
            self._vars = vars
            self.unknown_keys: set[str] = set()

        def __getitem__(self, key: str) -> str:
            if key not in ["tag", "prev_tag"] and key not in self._vars:
                self.unknown_keys.add(key)
            return ""

    mapping = _UnknownKeyDetector(link_templates.vars)
    link_templates.template.format_map(mapping)
    link_templates.template_first.format_map(mapping)
    link_templates.template_last.format_map(mapping)
    if mapping.unknown_keys:
        yuio.io.error(
            "Some variables used in link templates are missing "
            "from `release_link_template_vars`: `%s`. Please, add them to config",
            ", ".join(sorted(mapping.unknown_keys)),
        )
        raise yuio.app.AppError("Configuration is incorrect")


def _find_sections(version: FindMode | str, changelog: Changelog, config: Config):
    canonized_version = None
    if isinstance(version, str):
        canonized_version = (
            _canonize_version(_parse_version(version, config), config) or version
        )
    logger.debug(
        "searching for release %r, canonized_version=%r", version, canonized_version
    )
    is_latest = True
    for i, section in enumerate(changelog.sections):
        if section.is_unreleased():
            logger.debug("checking unreleased section")
            if version is FindMode.UNRELEASED:
                yield i, section, None
        elif release := section.as_release():
            logger.debug("checking release %r", release.version)
            if version is FindMode.LATEST or (
                release.canonized_version == canonized_version
            ):
                yield i, section, is_latest
                if version is FindMode.LATEST:
                    version = release.version
                    canonized_version = release.canonized_version
            is_latest = False


def _edit_section(changelog: Changelog, ctx: Context, section: Section):
    logger.debug("external edit for %s", section.what())
    to_edit = (
        "<!--\n"
        f" Edit changelog for {section.what()}.\n"
        " An empty changelog aborts the command.\n"
        "-->\n\n"
    )
    to_edit += _render(changelog, ctx, section.to_tokens(include_heading=False))
    to_edit = yuio.io.edit(to_edit, comment_marker=None, file_ext=".md")
    to_edit = _COMMENT_RE.sub("", to_edit).strip()
    if not to_edit:
        raise yuio.app.AppError("Got an empty changelog, command aborted")
    else:
        content = changelog.parser.parse(to_edit, changelog.parser_env)
        root = SyntaxTreeNode(content)
        subsections: list[SubSection] = []
        for subheading, content in _split_into_sections(root.children, 3):
            subsection = SubSection(heading=subheading, content=content)
            _detect_subsection_metadata(subsection, ctx)
            subsections.append(subsection)
        section.subsections = subsections
        logger.debug("external edit successful")


def _find_latest_version(
    changelog: Changelog, repo_versions: dict[str, RepoVersion] | None, config: Config
) -> tuple[Version | None, str | None, str | None]:
    logger.debug("trying to determine the latest release")
    max_version = None
    max_version_str = None
    max_version_str_canonized = None
    for section in changelog.sections:
        release = section.as_release()
        if not release:
            continue
        if release.parsed_version is None:
            _report_latest_version_fail("release", release.version, release.map, config)
        logger.debug("checking release version %r", release.canonized_version)
        if max_version is None or release.parsed_version > max_version:
            max_version = release.parsed_version
            max_version_str = release.version
            max_version_str_canonized = release.canonized_version
    if repo_versions is not None:
        for data in repo_versions.values():
            if data.parsed_version is None:
                _report_latest_version_fail("release", data.version, None, config)
            logger.debug("checking tag %r", data.canonized_version)
            if max_version is None or data.parsed_version > max_version:
                max_version = data.parsed_version
                max_version_str = data.version
                max_version_str_canonized = data.canonized_version
    logger.debug("latest release is %r", max_version_str)
    return max_version, max_version_str, max_version_str_canonized


def _report_latest_version_fail(
    what: str, version: str | None, map: tuple[int, int] | None, config: Config
) -> _t.Never:
    if config.version_format is VersionFormat.NONE:
        reason = "`version_format` is set to `none`"
        args = ()
    else:
        reason = f"{what} `%s` "
        if map:
            reason += f"on line `{map[0] + 1}` "
        reason += "does not follow %s specification"
        args = (version, config.version_format.value)
    raise yuio.app.AppError(
        "Can't determine the latest version because " + reason, *args
    )


_PY_TO_SEMVER_PRE_RELEASE = {
    "a": "alpha",
    "alpha": "alpha",
    "b": "beta",
    "beta": "beta",
    "rc": "rc",
}


def _bump_version(
    version: Version,
    mode: BumpMode | None,
    alpha: bool,
    beta: bool,
    rc: bool,
    unreleased: Section,
    config: Config,
) -> str:
    if isinstance(version, packaging.version.Version):
        epoch = version.epoch
        major, minor, patch = tuple(
            version.release[i] if len(version.release) >= i else 0 for i in range(3)
        )
        pre = version.pre
        post = version.post
    elif isinstance(version, semver.Version):
        if mode is BumpMode.POST:
            raise yuio.app.AppError(
                "Semver schema doesn't support creating post releases",
            )
        epoch = 0
        major, minor, patch = version.major, version.minor, version.patch
        pre = _split_semver_prerelease(version.prerelease)
        post = None
    else:
        assert False
    if mode is BumpMode.AUTO:
        mode = _suggest_bump(unreleased, config)
    match mode:
        case BumpMode.MAJOR:
            major, minor, patch, pre, post = major + 1, 0, 0, None, None
        case BumpMode.MINOR:
            major, minor, patch, pre, post = major, minor + 1, 0, None, None
        case BumpMode.PATCH:
            major, minor, patch, pre, post = major, minor, patch + 1, None, None
        case BumpMode.POST:
            if pre is not None:
                raise yuio.app.AppError(
                    "Creating a post-release for a pre-release is probably "
                    "a mistake. Please, specify version manually if you're sure "
                    "you want to do this",
                    version,
                )
            if post is None:
                post = 0
            else:
                post += 1
        case BumpMode.AUTO:
            assert False
        case None:
            assert alpha or beta or rc
            if not pre:
                raise yuio.app.AppError(
                    "Can't create a pre-release without bumping "
                    "a primary version component: latest release `%s` "
                    "is not a pre-release",
                    version,
                )

    if (alpha or beta or rc) and pre and pre[0] not in _PY_TO_SEMVER_PRE_RELEASE:
        raise yuio.app.AppError(
            "Can't create a pre-release: latest release `%s` has "
            "unknown label `%r`. Please, specify version manually",
            version,
            pre[0],
        )
    elif alpha:
        if pre and pre[0] in ("a", "alpha"):
            pre = ("a", pre[1] + 1)
        elif pre:
            if pre[0] == "rc":
                what = "a release candidate"
            else:
                what = "a beta pre-release"
            raise yuio.app.AppError("Can't create an alpha pre-release after %s", what)
        else:
            pre = ("a", 0)
    elif beta:
        if pre and pre[0] in ("b", "beta"):
            pre = ("b", pre[1] + 1)
        elif pre and pre[0] == "rc":
            raise yuio.app.AppError(
                f"Can't create a beta pre-release after a release candidate"
            )
        else:
            pre = ("b", 0)
    elif rc:
        if pre and pre[0] == "rc":
            pre = ("rc", pre[1] + 1)
        else:
            pre = ("rc", 0)

    if isinstance(version, packaging.version.Version):
        if epoch > 0:
            prefix = f"{epoch}!"
        else:
            prefix = ""
        suffix = ""
        if pre:
            suffix += "".join(map(str, pre))
        if post is not None:
            suffix += f".post{post}"
        bumped = packaging.version.Version(f"{prefix}{major}.{minor}.{patch}{suffix}")
    else:
        suffix = ""
        if pre:
            pre_str = f"{_PY_TO_SEMVER_PRE_RELEASE[pre[0]]}{pre[1]}"
        else:
            pre_str = None
        bumped = semver.Version(major, minor, patch, pre_str)

    logger.debug(
        "performed version bump: mode=%r, alpha=%r, beta=%r, rc=%r, result=%r",
        mode,
        alpha,
        beta,
        rc,
        bumped,
    )

    canonical = _canonize_version(_t.cast(Version, bumped), config)
    assert canonical
    return canonical


_SEMVER_PRERELEASE_RE = re.compile(r"^([a-zA-Z0-9.-]*?)([0-9]*)$")


def _split_semver_prerelease(pre: str | None) -> tuple[str, int] | None:
    if pre is None:
        return None
    if match := _SEMVER_PRERELEASE_RE.match(pre):
        return match.group(1).rstrip(".-"), int(match.group(2))
    else:
        return pre, 0


def _suggest_bump(unreleased: Section, config: Config) -> BumpMode:
    logger.debug("checking change categories to determine appropriate version bump")
    has_major = has_minor = has_patch = False
    for subsection in unreleased.subsections:
        if subsection.category_kind != SubSectionCategoryKind.KNOWN:
            logger.debug(
                "skipping subsection %s: category unknown", subsection.category
            )
            continue
        logger.debug("inspecting subsection %s", subsection.category)
        if subsection.category in config.full_bump_major_categories:
            logger.debug("subsection warrants a major bump")
            has_major = True
        elif subsection.category in config.full_bump_minor_categories:
            logger.debug("subsection warrants a minor bump")
            has_minor = True
        elif subsection.category in config.full_bump_patch_categories:
            logger.debug("subsection warrants a patch bump")
            has_patch = True
    if has_major:
        return BumpMode.MAJOR
    elif has_minor:
        return BumpMode.MINOR
    elif has_patch:
        return BumpMode.PATCH
    else:
        raise yuio.app.AppError(
            "Can't determine which version component to bump: changelog for unreleased "
            "version doesn't have change categories from `bump_major_categories`, "
            "`bump_minor_categories`, or `bump_patch_categories`"
        )


def _is_pre_release(version: Version | None) -> bool | None:
    if version is None:
        return None
    elif isinstance(version, packaging.version.Version):
        return version.is_prerelease
    elif isinstance(version, semver.Version):
        return version.prerelease is not None
    else:
        assert False


def _is_post_release(version: Version | None) -> bool | None:
    if version is None:
        return None
    elif isinstance(version, packaging.version.Version):
        return version.is_postrelease
    elif isinstance(version, semver.Version):
        return False
    else:
        assert False


class Theme(yuio.theme.DefaultTheme):
    colors = {
        "code": "bold",
        "msg/text:report_error": ["red"],
        "msg/text:report_warning": ["yellow"],
        "msg/text:report_weak_warning": ["cyan"],
        "msg/text:report_info": ["cyan"],
    }


main.theme = Theme



"""
This type stub file was generated by pyright.
"""

class GitUrlParsed:
    def __init__(self, parsed_info): ...
    @property
    def valid(self) -> bool: ...
    @property
    def host(self) -> str: ...
    @property
    def resource(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def user(self) -> str: ...
    @property
    def pathname(self) -> str: ...
    @property
    def platform(self) -> str: ...

def parse(url: str, check_domain: bool = ...) -> GitUrlParsed: ...
def validate(url: str, check_domain: bool = ...) -> bool: ...



from __future__ import annotations

import enum
import pathlib
import typing as _t

import yuio.io

from cl_keeper.config import Config, IssueCode, IssueSeverity, LinkTemplates
from cl_keeper.typing import SupportsPos


class IssueScope(enum.Enum):
    """
    Scope where an issue happened.

    """

    #: For issues originating from environment,
    #: i.e. repository tags not being properly formatted.
    EXTERNAL = 0

    #: For issues originating from the changelog file.
    CHANGELOG = 1


#: Position of an issue, a pair of 0-based line numbers, first and last line.
IssuePosition: _t.TypeAlias = tuple[int, int]


class Context:
    def __init__(
        self,
        path: pathlib.Path,
        src: str,
        config: Config,
        strict: bool,
        link_templates: LinkTemplates,
    ) -> None:
        self.path = path
        self.src = src
        self.lines = self.src.splitlines()
        self.config = config
        self.strict = strict
        self.link_templates = link_templates
        self._has_errors = False
        self._messages: list[
            tuple[
                str,
                tuple[_t.Any, ...],
                IssuePosition | None,
                IssueCode,
                IssueScope,
                IssueSeverity,
            ]
        ] = []

    def reset(self, src: str):
        self.src = src
        self.lines = self.src.splitlines()
        self._has_errors = False
        self._messages = []

    def issue(
        self,
        code: IssueCode,
        msg: str,
        *args: _t.Any,
        pos: int | tuple[int, int] | SupportsPos | None = None,
        scope: IssueScope = IssueScope.CHANGELOG,
    ):
        severity = self.config.severity.get(code, code.default_severity())
        severity = self._up(severity)
        if severity is IssueSeverity.NONE:
            return
        if severity.value >= IssueSeverity.ERROR.value:
            self._has_errors = True
        if isinstance(pos, SupportsPos):
            pos = pos.map
        elif isinstance(pos, int):
            pos = (pos, pos + 1)
        self._messages.append((msg, args, pos, code, scope, severity))

    def report(self):
        self._messages.sort(key=lambda x: (x[3].value, -x[4].value, x[2] or (0, 0)))
        prev_pos = None
        prev_title = None
        for i, (msg, args, pos, code, _, severity) in enumerate(self._messages):
            if i >= 50:
                self._print_source(prev_pos)
                skipped = len(self._messages) - 50
                yuio.io.error(
                    "<c b>+ %s more message%s skipped.</c>",
                    skipped,
                    "" if skipped == 1 else "s",
                )
                return

            color, title = self._color_and_title(severity)
            if not color or not title:
                continue
            msg += f" [{code.value}]"
            if pos != prev_pos:
                self._print_source(prev_pos)
            if pos and (pos != prev_pos or title != prev_title):
                yuio.io.info(f"<c b>{title} on line {pos[0] + 1}:</c>", color=color)
            if pos:
                yuio.io.info("  - " + msg, *args, color=color)
            else:
                yuio.io.info(f"<c b>{title}: {msg}</c>", *args, color=color)
            prev_pos = pos
            prev_title = title
        self._print_source(prev_pos)

    def has_errors(self) -> bool:
        return self._has_errors

    def has_messages(self) -> bool:
        return bool(self._messages)

    def exit_if_has_errors(self):
        if self.has_errors():
            exit(3)

    def _print_source(self, pos: tuple[int, int] | None):
        if pos:
            for line in range(*pos):
                if line < len(self.lines):
                    yuio.io.info(
                        "  <c dim>%4d |</c> %s",
                        line + 1,
                        self.lines[line],
                    )
                else:
                    break

    def _up(self, severity: IssueSeverity) -> IssueSeverity:
        return IssueSeverity(
            min(severity.value + self.strict, IssueSeverity.ERROR.value)
        )

    @staticmethod
    def _color_and_title(severity: IssueSeverity):
        match severity:
            case IssueSeverity.ERROR:
                return "report_error", "Error"
            case IssueSeverity.WARNING:
                return "report_warning", "Warning"
            case IssueSeverity.WEAK_WARNING:
                return "report_weak_warning", "Weak warning"
            case IssueSeverity.INFO:
                return "report_info", "Info"
            case IssueSeverity.NONE:
                return None, None



import difflib
import pathlib

import yuio.io
import yuio.md
from markdown_it.token import Token

from cl_keeper.config import Wrapping
from cl_keeper.context import Context
from cl_keeper.model import Changelog


def render(
    changelog: Changelog,
    ctx: Context,
    tokens: list[Token] | None = None,
    disable_wrapping: bool = False,
) -> str:
    if tokens is None:
        tokens = changelog.to_tokens()
    parser = changelog.parser
    if disable_wrapping:
        parser.options["mdformat"]["wrap"] = "no"
    elif isinstance(ctx.config.format_wrapping, Wrapping):
        parser.options["mdformat"]["wrap"] = ctx.config.format_wrapping.value
    else:
        parser.options["mdformat"]["wrap"] = ctx.config.format_wrapping
    parser.options["mdformat"]["number"] = True
    rendered = parser.renderer.render(tokens, parser.options, changelog.parser_env)
    return parser.render(rendered)


def print_diff(l: str, r: str, path: pathlib.Path):
    diff = "".join(
        difflib.unified_diff(
            l.splitlines(keepends=True),
            r.splitlines(keepends=True),
            fromfile=f"{path}:before",
            tofile=f"{path}:after",
        )
    )

    yuio.io.heading("Diff")

    if not diff:
        yuio.io.info("Diff is empty")
    else:
        yuio.io.raw(
            yuio.md.SyntaxHighlighter.get_highlighter("diff").highlight(
                yuio.io.get_theme(),
                diff,
            )
        )



import pathlib
import textwrap

import pytest
from packaging.version import Version as PyVersion
from semver import Version as SemverVersion

from cl_keeper.config import Config, VersionFormat
from cl_keeper.context import Context
from cl_keeper.model import Section, SubSection
from cl_keeper.parse import canonize_version, parse, parse_version


@pytest.mark.parametrize(
    "format,version,expected",
    [
        (VersionFormat.NONE, "whatever", None),
        (
            VersionFormat.SEMVER,
            "1.0.0-beta.0+12345",
            SemverVersion(1, 0, 0, "beta.0", "12345"),
        ),
        (VersionFormat.SEMVER_STRICT, "1.0.0", SemverVersion(1, 0, 0)),
        (VersionFormat.SEMVER_STRICT, "1.0.0-alpha1", SemverVersion(1, 0, 0, "alpha1")),
        (VersionFormat.SEMVER_STRICT, "1.0.0-beta1", SemverVersion(1, 0, 0, "beta1")),
        (VersionFormat.SEMVER_STRICT, "1.0.0-rc1", SemverVersion(1, 0, 0, "rc1")),
        (VersionFormat.SEMVER_STRICT, "1.0", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0.0", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0-beta.1", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0-foo1", None),
        (VersionFormat.PYTHON, "1.0.0", PyVersion("1.0.0")),
        (VersionFormat.PYTHON, "1.0.0b0", PyVersion("1.0.0b0")),
        (VersionFormat.PYTHON, "1.0.0-beta.0", PyVersion("1.0.0b0")),
        (VersionFormat.PYTHON, "1.0.0.post0+foo", PyVersion("1.0.0.post0+foo")),
        (VersionFormat.PYTHON, "2!1.0.0", PyVersion("2!1.0.0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0", PyVersion("1.0.0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0a0", PyVersion("1.0.0a0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0b0", PyVersion("1.0.0b0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0rc0", PyVersion("1.0.0rc0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0.post0", PyVersion("1.0.0.post0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0a0.post0", PyVersion("1.0.0a0.post0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0b0.post0", PyVersion("1.0.0b0.post0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0rc0.post0", PyVersion("1.0.0rc0.post0")),
        (VersionFormat.PYTHON_STRICT, "2!1.0.0", PyVersion("2!1.0.0")),
        (VersionFormat.PYTHON_STRICT, "1.0.0-a0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0-alpha0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0a.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0alpha.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0-post0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0post.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0.post-0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0", PyVersion("1.0.0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-alpha0", PyVersion("1.0.0a0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-beta0", PyVersion("1.0.0b0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-rc0", PyVersion("1.0.0rc0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-post0", PyVersion("1.0.0.post0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-alpha0.post0", PyVersion("1.0.0a0.post0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-beta0.post0", PyVersion("1.0.0b0.post0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-rc0.post0", PyVersion("1.0.0rc0.post0")),
        (VersionFormat.PYTHON_SEMVER, "2!1.0.0", PyVersion("2!1.0.0")),
        (VersionFormat.PYTHON_SEMVER, "1.0.0a0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0b0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0rc0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0a0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0b0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0rc0.post0", None),
    ],
)
def test_parse_version(format, version, expected):
    assert parse_version(version, Config(version_format=format)) == expected


@pytest.mark.parametrize(
    "format,version,expected",
    [
        (VersionFormat.NONE, "whatever", None),
        (VersionFormat.SEMVER, "1.0.0-beta.0+12345", "1.0.0-beta.0+12345"),
        (VersionFormat.SEMVER_STRICT, "1.0.0", "1.0.0"),
        (VersionFormat.SEMVER_STRICT, "1.0.0-alpha1", "1.0.0-alpha1"),
        (VersionFormat.SEMVER_STRICT, "1.0.0-beta1", "1.0.0-beta1"),
        (VersionFormat.SEMVER_STRICT, "1.0.0-rc1", "1.0.0-rc1"),
        (VersionFormat.SEMVER_STRICT, "1.0", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0.0", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0-beta.1", None),
        (VersionFormat.SEMVER_STRICT, "1.0.0-foo1", None),
        (VersionFormat.PYTHON, "1.0.0", "1.0.0"),
        (VersionFormat.PYTHON, "1.0.0b0", "1.0.0b0"),
        (VersionFormat.PYTHON, "1.0.0-beta.0", "1.0.0b0"),
        (VersionFormat.PYTHON, "1.0.0-beta.post0", "1.0.0b0.post0"),
        (VersionFormat.PYTHON, "2!1.0.0", "2!1.0.0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0", "1.0.0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0a0", "1.0.0a0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0b0", "1.0.0b0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0rc0", "1.0.0rc0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0.post0", "1.0.0.post0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0a0.post0", "1.0.0a0.post0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0b0.post0", "1.0.0b0.post0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0rc0.post0", "1.0.0rc0.post0"),
        (VersionFormat.PYTHON_STRICT, "2!1.0.0", "2!1.0.0"),
        (VersionFormat.PYTHON_STRICT, "1.0.0-a0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0-alpha0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0a.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0alpha.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0-post0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0post.0", None),
        (VersionFormat.PYTHON_STRICT, "1.0.0.post-0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0", "1.0.0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-alpha0", "1.0.0-alpha0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-beta0", "1.0.0-beta0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-rc0", "1.0.0-rc0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-post0", "1.0.0-post0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-alpha0.post0", "1.0.0-alpha0.post0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-beta0.post0", "1.0.0-beta0.post0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0-rc0.post0", "1.0.0-rc0.post0"),
        (VersionFormat.PYTHON_SEMVER, "2!1.0.0", "2!1.0.0"),
        (VersionFormat.PYTHON_SEMVER, "2!1.0.0-beta0", "2!1.0.0-beta0"),
        (VersionFormat.PYTHON_SEMVER, "1.0.0a0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0b0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0rc0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0a0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0b0.post0", None),
        (VersionFormat.PYTHON_SEMVER, "1.0.0rc0.post0", None),
    ],
)
def test_canonize(format, version, expected):
    config = Config(version_format=format)
    assert canonize_version(parse_version(version, config), config) == expected


@pytest.mark.parametrize(
    "input,config",
    [
        pytest.param(
            """
            """,
            None,
            id="empty",
        ),
        pytest.param(
            """
            """,
            None,
            id="only-heading",
        ),
        pytest.param(
            """
            ## Unreleased
            """,
            None,
            id="unreleased-simple",
        ),
        pytest.param(
            """
            ## [Unreleased]
            """,
            None,
            id="unreleased-unlinked",
        ),
        pytest.param(
            """
            ## [Unreleased]

            [Unreleased]: https://example.com/
            """,
            None,
            id="unreleased-linked",
        ),
        pytest.param(
            """
            ## something-something unreleased, maybe?
            """,
            None,
            id="unreleased-broken",
        ),
        pytest.param(
            """
            ## Unreleased
            ### Added
            - a
            - b
            ### Removed
            - c
            - d
            ### Unknown
            - e
            - f
            """,
            None,
            id="unreleased-subsections",
        ),
        pytest.param(
            """
            ## 1.0.0
            """,
            None,
            id="release",
        ),
        pytest.param(
            """
            ## v1.0.0
            """,
            None,
            id="release-with-prefix",
        ),
        pytest.param(
            """
            ## [1.0.0]
            """,
            None,
            id="release-unlinked",
        ),
        pytest.param(
            """
            ## [1.0.0] - 2020-01-01
            """,
            None,
            id="release-unlinked-with-date",
        ),
        pytest.param(
            """
            ## 2020-01-01 [1.0.0]
            """,
            None,
            id="release-unlinked-with-date-before-version",
        ),
        pytest.param(
            """
            ## [1.0.0]
            [1.0.0]: https://example.com/
            """,
            None,
            id="release-linked",
        ),
        pytest.param(
            """
            ## ~ 1.0.0 ~ 2020-01-01
            """,
            None,
            id="release-with-decorations-error",
        ),
        pytest.param(
            """
            ## ~ 1.0.0 ~ 2020-01-01
            """,
            Config(
                version_decorations=("~ ", " ~"),
                release_date_decorations=(" ", ""),
            ),
            id="release-with-decorations-ok",
        ),
        pytest.param(
            """
            ## 1.0.0 - 2020-01-01
            """,
            None,
            id="release-with-date",
        ),
        pytest.param(
            """
            ## 1.0.0 - 2020-20-20
            """,
            None,
            id="release-with-broken-date",
        ),
        pytest.param(
            """
            ## 2020-01-01 - 1.0.0
            """,
            None,
            id="release-with-date-before-version",
        ),
        pytest.param(
            """
            ## 1.0.0 - [YANKED]
            """,
            None,
            id="release-with-postfix",
        ),
        pytest.param(
            """
            ## 1.0.0 - 2020-01-01 - [YANKED]
            """,
            None,
            id="release-with-date-and-postfix",
        ),
        pytest.param(
            """
            ## 2020-01-01 - 1.0.0 - [YANKED]
            """,
            None,
            id="release-with-date-before-version-and-postfix",
        ),
        pytest.param(
            """
            ## 1.0.0 - (2020-01-01) - [YANKED]
            """,
            None,
            id="release-with-date-and-postfix-braces",
        ),
        pytest.param(
            """
            ## 2020-01-01 - (1.0.0) - [YANKED]
            """,
            None,
            id="release-with-date-before-version-and-postfix-braces",
        ),
        pytest.param(
            """
            # Changelog

            ## [Unreleased]

            ## [1.0.0] - 2020-01-01

            ### Added

            - foo
            - bar

            ### Removed

            - baz

            ## [0.0.0-beta3]

            - Beta pre-release.

            [unreleased]: https://example.com/unreleased
            [1.0.0]: https://example.com/1.0.0
            [0.0.0-beta3]: https://example.com/0.0.0-beta3
            """,
            None,
            id="multiple-releases-with-sections",
        ),
        pytest.param(
            """
            # Changelog

            ## [Unreleased]

            ## [1.0.0] - 2020-01-01

            ### Added

            - foo
            - bar

            # What??

            ### Removed

            - baz

            ## [0.0.0-beta3]

            - Beta pre-release.

            [unreleased]: https://example.com/unreleased
            [1.0.0]: https://example.com/1.0.0
            [0.0.0-beta3]: https://example.com/0.0.0-beta3
            """,
            None,
            id="first-level-heading-in-the-middle",
        ),
        pytest.param(
            """
            ### Added
            - foo
            """,
            None,
            id="subsection-not-attached",
        ),
        pytest.param(
            """
            ## 1.0.0
            ### ~ added ~
            - foo
            """,
            None,
            id="subsection-format-error",
        ),
        pytest.param(
            """
            ## 1.0.0
            ### ~ added ~
            - foo
            """,
            Config(extra_change_categories={"added": "~ added ~"}),
            id="subsection-format-ok",
        ),
        pytest.param(
            """
            ## who is this?
            It's a me, Mario!
            """,
            None,
            id="unrecognized-section",
        ),
        pytest.param(
            """
            ## 1.0.0.b10
            """,
            None,
            id="version-canonization-fail",
        ),
        pytest.param(
            """
            ## 1.0.0.b10
            """,
            Config(version_format=VersionFormat.PYTHON),
            id="version-canonization-ok",
        ),
        pytest.param(
            """
            ## 1.0.0.b10
            """,
            Config(version_format=VersionFormat.NONE),
            id="version-canonization-none",
        ),
    ],
)
def test_parse(input: str, config, data_regression):
    input = textwrap.dedent(input).strip()
    ctx = Context(
        pathlib.Path("__test__"),
        input,
        config or Config(),
        False,
        None,  # type: ignore
    )

    changelog = parse(ctx)

    data = dict(
        sections=[_serialize_section(section) for section in changelog.sections],
        messages=_serialize_messages(ctx),
    )

    data_regression.check(data)


def _serialize_section(section: Section):
    subsections = [
        _serialize_subsection(subsection) for subsection in section.subsections
    ]
    if release := section.as_release():
        return dict(
            type="RELEASE",
            version=release.version,
            parsed_version=release.parsed_version and str(release.parsed_version),
            canonized_version=release.canonized_version,
            version_link=release.version_link,
            version_label=release.version_label,
            release_date=release.release_date and release.release_date.isoformat(),
            release_date_fmt=release.release_date_fmt,
            release_comment=release.release_comment,
            subsections=subsections,
        )
    elif unreleased := section.as_unreleased():
        return dict(
            type="UNRELEASED",
            version_link=unreleased.version_link,
            version_label=unreleased.version_label,
            subsections=subsections,
        )
    else:
        return dict(
            type="TRIVIA",
            subsections=subsections,
        )


def _serialize_subsection(subsection: SubSection):
    return dict(
        type=subsection.type.value,
        category_kind=subsection.category_kind.value,
        category=subsection.category,
        sort_key=subsection.sort_key,
    )


def _serialize_messages(ctx: Context):
    return [
        dict(
            mag=msg % args if args else msg,
            pos=pos,
            code=code.value,
            scope=scope.value,
            severity=severity.value,
        )
        for msg, args, pos, code, scope, severity in ctx._messages  # type: ignore
    ]



from __future__ import annotations

import abc
import typing as _t


@_t.runtime_checkable
class SupportsPos(_t.Protocol):
    @property
    @abc.abstractmethod
    def map(self) -> tuple[int, int] | None:
        """
        Return object's position in the changelog file.

        """


class Ord(_t.Protocol):
    @abc.abstractmethod
    def __lt__(self, other: _t.Self, /) -> bool: ...



from __future__ import annotations

import dataclasses
import datetime
import enum
import typing as _t
from dataclasses import dataclass

from markdown_it import MarkdownIt
from markdown_it.token import Token
from markdown_it.tree import SyntaxTreeNode


class Version:
    def __lt__(self, other: _t.Self, /) -> bool: ...
    def __le__(self, other: _t.Self, /) -> bool: ...
    def __gt__(self, other: _t.Self, /) -> bool: ...
    def __ge__(self, other: _t.Self, /) -> bool: ...
    def __eq__(self, other: object, /) -> bool: ...
    def __ne__(self, other: object, /) -> bool: ...


@dataclass(kw_only=True)
class RepoVersion:
    version: str
    parsed_version: Version | None
    canonized_version: str
    author_date: datetime.date
    committer_date: datetime.date


@dataclass(kw_only=True)
class Changelog:
    sections: list[Section]
    parser: MarkdownIt
    parser_env: dict[str, _t.Any]

    def to_tokens(self) -> list[Token]:
        return [token for section in self.sections for token in section.to_tokens()]


@dataclass(kw_only=True)
class Section:
    """
    A single section of a changelog file.

    """

    #: Section heading, can be missing for trivia sections.
    heading: SyntaxTreeNode | None

    #: Section content.
    subsections: list[SubSection]

    def what(self) -> str:
        return "trivia section"

    def is_trivia(self) -> bool:
        return not isinstance(self, (ReleaseSection, UnreleasedSection))

    def is_release(self) -> bool:
        return isinstance(self, ReleaseSection)

    def as_release(self) -> ReleaseSection | None:
        if isinstance(self, ReleaseSection):
            return self
        else:
            return None

    def is_unreleased(self) -> bool:
        return isinstance(self, UnreleasedSection)

    def as_unreleased(self) -> UnreleasedSection | None:
        if isinstance(self, UnreleasedSection):
            return self
        else:
            return None

    @property
    def map(self) -> tuple[int, int] | None:
        if self.heading and self.heading.map:
            return self.heading.map
        for child in self.subsections:
            if child.map:
                return child.map
        return None

    def to_tokens(self, include_heading: bool = True) -> list[Token]:
        tokens: list[Token] = []
        if self.heading and include_heading:
            tokens.extend(self.heading.to_tokens())
        for child in self.subsections:
            tokens.extend(child.to_tokens())
        return tokens

    def walk(self):
        if self.heading is not None:
            yield from self.heading.walk()
        for subsection in self.subsections:
            yield from subsection.walk()


@dataclass(kw_only=True)
class UnreleasedSection(Section):
    #: Link attached to version string, if there was any.
    version_link: str | None

    #: Link label attached to version string, if there was any.
    version_label: str | None

    def what(self) -> str:
        return "unreleased section"


@dataclass(kw_only=True)
class ReleaseSection(Section):
    #: Version string extracted from the heading, only set in release sections.
    version: str

    #: Parsed version suitable for ordering, can be absent if `version` failed to parse.
    parsed_version: Version | None

    #: Canonized version suitable for comparing for equality.
    canonized_version: str

    #: Link attached to version string, if there was any.
    version_link: str | None

    #: Link label attached to version string, if there was any.
    version_label: str | None

    #: Known release date, either extracted from heading or from git metadata.
    release_date: datetime.date | None

    #: Formatted release date, possible an invalid one.
    release_date_fmt: str | None

    #: Heading text that was found after release version and date.
    release_comment: str | None

    def what(self) -> str:
        return f"release `{self.version}`"


class SubSectionType(enum.Enum):
    #: Subsection that contains some text that doesn't describe changes.
    TRIVIA = "TRIVIA"

    #: Subsection that contains release changes.
    CHANGES = "CHANGES"


class SubSectionCategoryKind(enum.Enum):
    #: Subsection's category was recognized.
    KNOWN = "KNOWN"

    #: Subsection's category was not recognized.
    UNKNOWN = "UNKNOWN"


@dataclass(kw_only=True)
class SubSection:
    """
    Content of a changelog section.

    """

    #: Subsection type.
    type: SubSectionType = SubSectionType.TRIVIA

    #: Subsection heading, can be missing for trivia subsections.
    heading: SyntaxTreeNode | None = None

    #: Whether `category` was recognized as a known category. Trivia subsections
    #: are always `UNKNOWN`.
    category_kind: SubSectionCategoryKind = SubSectionCategoryKind.UNKNOWN

    #: Category of changes, recognized by scanning heading
    #: with `change_categories_map`. For `UNKNOWN` category kinds, this contains
    #: case-folded heading content for merging subsections with same headings.
    #: For trivia subsections, this contains an empty string.
    category: str = ""

    #: Index of subsection category in the canonical order of subsections.
    sort_key: int | None = None

    #: Subsection content.
    content: list[SyntaxTreeNode] = dataclasses.field(default_factory=list)

    @property
    def map(self) -> tuple[int, int] | None:
        if self.heading and self.heading.map:
            return self.heading.map
        for child in self.content:
            if child.map:
                return child.map
        return None

    def to_tokens(self) -> list[Token]:
        tokens: list[Token] = []
        if self.heading:
            tokens.extend(self.heading.to_tokens())
        for child in self.content:
            tokens.extend(child.to_tokens())
        return tokens

    def walk(self):
        for node in self.content:
            yield from node.walk()



from __future__ import annotations

import enum
import functools
import pathlib
import typing as _t
from dataclasses import dataclass

import yuio.app
import yuio.config
import yuio.parse


class Wrapping(enum.Enum):
    NO = "no"
    """
    Disable wrapping, print all paragraphs in a single line.

    """

    KEEP = "keep"
    """
    Keep the original file wrapping.

    """


class IssueSeverity(enum.Enum):
    """
    Severity of the issue.

    """

    ERROR = 4
    """
    Error.

    """

    WARNING = 3
    """
    Warning.

    """

    WEAK_WARNING = 2
    """
    Weak warning.

    """

    INFO = 1
    """
    Info.

    """

    NONE = 0
    """
    Not displayed.

    """


class IssueCode(enum.Enum):
    GENERAL_FORMATTING_ERROR = "GeneralFormattingError"
    """
    A formatting error related to the general structure of the changelog.

    """

    INVALID_VERSION = "InvalidVersion"
    """
    Release version doesn't follow the requested specification.

    """

    INVALID_TAG = "InvalidTag"
    """
    Tag format doesn't follow the requested versioning specification.

    To disable all repository checks, set :attr:`~Config.check_repo_tags` to ``False``.
    """

    MISSING_RELEASE_FOR_TAG = "MissingReleaseForTag"
    """
    Found a tag that has no associated release.

    To suppress this error for all versions before the given one,
    set `ignore_missing_releases_before`. To disable all repository checks,
    set :attr:`~Config.check_repo_tags` to ``False``.

    """

    MISSING_TAG_FOR_RELEASE = "MissingTagForRelease"
    """
    Found a release that has no associated tag in the repository.

    To disable all repository checks, set :attr:`~Config.check_repo_tags` to ``False``.

    """

    RELEASE_ORDERING = "ReleaseOrdering"
    """
    Releases are not ordered according to their versions.

    Releases are compared according to the chosen versioning specification.

    """

    UNRELEASED_HEADING_FORMAT = "UnreleasedHeadingFormat"
    """
    Heading for unreleased changes doesn't meet formatting requirements.

    Heading format is controlled by `unreleased_name`.

    """

    RELEASE_HEADING_FORMAT = "ReleaseHeadingFormat"
    """
    Heading for a release doesn't meet formatting requirements.

    """

    INVALID_RELEASE_DATE = "InvalidReleaseDate"
    """
    Release date can't be parsed.

    This usually indicates that the release date is incorrect. For example,
    it refers to a day that doesn't exist in the calendar.

    """

    MISSING_RELEASE_DATE = "MissingReleaseDate"
    """
    Release date is missing.

    :attr:`~Config.add_release_date` is ``True``, but release date is missing.

    """

    UNEXPECTED_RELEASE_DATE = "UnexpectedReleaseDate"
    """
    Release date should not be present.

    :attr:`~Config.add_release_date` is ``False``, but release date is present.

    """

    MISSING_RELEASE_LINK = "MissingReleaseLink"
    """
    Release link is missing.

    :attr:`~Config.add_release_link` is ``True``, but release link is missing.

    """

    UNEXPECTED_RELEASE_LINK = "UnexpectedReleaseLink"
    """
    Release link should not be present.

    :attr:`~Config.add_release_link` is ``False``, but release link is present.

    """

    INCORRECT_RELEASE_DATE = "IncorrectReleaseDate"
    """
    Release date doesn't match date of the release commit.

    To disable all repository checks, set :attr:`~Config.check_repo_tags` to ``False``.

    """

    INCORRECT_RELEASE_LINK = "IncorrectReleaseLink"
    """
    Release link is incorrect.

    Release links are generated using `release_link_preset`
    and `release_link_template`. We use order of releases from the changelog
    to generate links for commit ranges.

    """

    DUPLICATE_RELEASES = "DuplicateReleases"
    """
    Found multiple releases for the same version.

    """

    EMPTY_RELEASE = "EmptyRelease"
    """
    Section for a release is empty.

    To ignore empty sections prior to some version,
    set `ignore_missing_releases_before`.

    """

    CHANGE_CATEGORY_ORDERING = "ChangeCategoryOrdering"
    """
    Order of sub-sections within a release doesn't match
    the one given in `change_categories`.

    """

    CHANGE_CATEGORY_HEADING_FORMAT = "ChangeCategoryHeadingFormat"
    """
    Heading for a change category doesn't meet formatting requirements.

    """

    UNKNOWN_CHANGE_CATEGORY = "UnknownChangeCategory"
    """
    Heading for a change category doesn't appear in `change_categories_map`.

    """

    DUPLICATE_CHANGE_CATEGORIES = "DuplicateChangeCategories"
    """
    Found multiple release sub-sections with the same title.

    """

    EMPTY_CHANGE_CATEGORY = "EmptyChangeCategory"
    """
    Found a release sub-section without content.

    """

    def default_severity(self) -> IssueSeverity:
        match self:
            case self.INVALID_VERSION:
                return IssueSeverity.ERROR
            case self.MISSING_RELEASE_FOR_TAG:
                return IssueSeverity.WEAK_WARNING
            case self.MISSING_TAG_FOR_RELEASE:
                return IssueSeverity.WEAK_WARNING
            case self.RELEASE_ORDERING:
                return IssueSeverity.ERROR
            case self.INVALID_RELEASE_DATE:
                return IssueSeverity.ERROR
            case self.DUPLICATE_RELEASES:
                return IssueSeverity.ERROR
            case self.EMPTY_RELEASE:
                return IssueSeverity.WARNING
            case self.INCORRECT_RELEASE_DATE:
                return IssueSeverity.WEAK_WARNING
            case self.DUPLICATE_CHANGE_CATEGORIES:
                return IssueSeverity.ERROR
            case self.EMPTY_CHANGE_CATEGORY:
                return IssueSeverity.ERROR
            case _:
                return IssueSeverity.WARNING


@dataclass
class LinkTemplates:
    template: str
    template_last: str
    template_first: str
    vars: dict[str, str]

    def has_unresolved_links(self):
        return not self.template or not self.template_last or not self.template_first

    def update(self, rhs: LinkTemplates):
        self.template = self.template or rhs.template
        self.template_last = self.template_last or rhs.template_last
        self.template_first = self.template_first or rhs.template_first
        self.update_vars(rhs.vars)
        return self

    def update_vars(self, vars: dict[str, str], override: bool = False):
        if override:
            self.vars.update(vars)
        else:
            for k, v in vars.items():
                self.vars.setdefault(k, v)
        return self


class ReleaseLinkPreset(enum.Enum):
    GITHUB = "github"
    """
    Will link releases to tags on github.

    To use this preset, add the ``repo`` template variable
    to :attr:`~Config.release_link_template_vars`.
    It should contain url of your repository.

    If you're using self-hosted GitHub, you can also add the ``host``
    template variable.

    **Example configuration:**

    .. code-block:: yaml

        release_link_preset: github
        release_link_template_vars:
            repo: "taminomara/cl-keeper"

    """

    GITLAB = "gitlab"
    """
    Will link releases to tags on gitlab.

    To use this preset, add the ``repo`` template variable
    to :attr:`~Config.release_link_template_vars`.
    It should contain url of your repository.

    If you're using self-hosted GitLab, you can also add the ``host``
    template variable.

    **Example configuration:**

    .. code-block:: yaml

        release_link_preset: gitlab
        release_link_template_vars:
            repo: "taminomara/cl-keeper"

    """

    NONE = "none"
    """
    Will not generate links automatically.

    """

    def get_links(self) -> LinkTemplates:
        match self:
            case ReleaseLinkPreset.GITHUB:
                return LinkTemplates(
                    "https://{host}/{repo}/compare/{prev_tag}...{tag}",
                    "https://{host}/{repo}/compare/{prev_tag}...HEAD",
                    "https://{host}/{repo}/releases/tag/{tag}",
                    {"host": "github.com"},
                )
            case ReleaseLinkPreset.GITLAB:
                return LinkTemplates(
                    "https://{host}/{repo}/-/compare/{prev_tag}...{tag}",
                    "https://{host}/{repo}/-/compare/{prev_tag}...HEAD",
                    "https://{host}/{repo}/-/tags/{tag}",
                    {"host": "gitlab.com"},
                )
            case ReleaseLinkPreset.NONE:
                return LinkTemplates("", "", "", {})


class VersionFormat(enum.Enum):
    """
    Specification for versioning formats.

    Different settings affect parsing versions and sorting changelog entries.

    .. list-table:: Syntax overview
        :header-rows: 1

        * - Schema
          - Release
          - Pre-release
          - Post-release
          - Post-release of a pre-release

        * - :attr:`~VersionFormat.SEMVER`
          - ``1.0.0``
          - ``1.0.0-b``, ``1.0.0-b0``, ``1.0.0-beta0``, etc.
          - not supported
          - not supported
        * - :attr:`~VersionFormat.SEMVER_STRICT`
          - ``1.0.0``
          - ``1.0.0-beta0``
          - not supported
          - not supported
        * - :attr:`~VersionFormat.PYTHON`
          - ``1.0.0``
          - ``1.0.0b0``, ``1.0.0-beta.0``, etc.
          - ``1.0.0post0``, ``1.0.0-rev.0``, etc.
          - ``1.0.0b1.post2``, ``1.0.0-beta.1post2``, etc.
        * - :attr:`~VersionFormat.PYTHON_STRICT`
          - ``1.0.0``
          - ``1.0.0b0``
          - ``1.0.0post0``
          - ``1.0.0b0.post0``
        * - :attr:`~VersionFormat.PYTHON_SEMVER`
          - ``1.0.0``
          - ``1.0.0-beta0``
          - ``1.0.0-post0``
          - ``1.0.0-beta0.post0``

    """

    SEMVER = "semver"
    """
    Semantic versioning, as per `semver 2.0 specification`__.

    __ https://semver.org/

    """

    SEMVER_STRICT = "semver-strict"
    """
    Semantic versioning with stricter constraints.

    Versions must contain exactly three components (i.e. ``1.0`` is not valid,
    ``1.0.0`` is valid.) Only ``alpha``, ``beta``, and ``rc`` keywords
    followed by a number are allowed in pre-release section.

    This schema ensures that all versions and tags are normalized
    and can be checked for equality without parsing them.

    """

    PYTHON = "python"
    """
    Python versioning specification, as per `PyPA specification`__.

    __ https://packaging.python.org/en/latest/discussions/versioning/

    """

    PYTHON_STRICT = "python-strict"
    """
    Python versioning specification with stricter constraints.

    Versions must contain exactly three components (i.e. ``1.0`` is not valid,
    ``1.0.0`` is valid.) Leading zeroes are not allowed. Pre- release
    part must not be separated by point or dash, and only ``a``, ``b``,
    ``rc`` or ``post`` keywords are allowed.

    This schema ensures that all versions and tags are normalized
    and can be checked for equality without parsing them.

    """

    PYTHON_SEMVER = "python-semver"
    """
    Python versioning specification using :attr:`~VersionFormat.SEMVER_STRICT` syntax.

    This schema ensures that that all versions and tags are normalized,
    and other tools that rely on semver specification can work with them.

    .. note::

       Semver 2.0 doesn't specify post-release semantics. As such, we use Python's
       semantics and syntax: ``1.0.0-post0``, ``1.0.0-rc1.post2``, etc.

       This tool will sort post-releases after actual releases, while other tools
       might not work this way.

    """

    NONE = "none"
    """
    No versioning semantics are implied.

    Automatic sorting of releases is disabled. Canonization of versions when
    searching for release of merging duplicate releases is also disabled.

    """


class GlobalConfig(yuio.config.Config):
    config_path: pathlib.Path | None = None
    """
    Override default path to config.

    """

    file: pathlib.Path | None = None
    """
    Path to the changelog file.

    """

    strict: bool = False
    """
    Increase severity of all messages by one level.

    """


class Config(yuio.config.Config):
    file: pathlib.Path = yuio.config.field(
        default=pathlib.Path("CHANGELOG.md"),
        parser=yuio.parse.ExistingPath(),
        flags=yuio.DISABLED,
    )
    """
    Path to the changelog file, relative to this config location.

    """

    format_wrapping: (
        _t.Annotated[int, yuio.parse.Gt(0)]
        | _t.Annotated[Wrapping, yuio.parse.Enum(doc_inline=True)]
    ) = yuio.config.field(
        default=90,
        help="max line width for wrapping markdown files",
    )
    """
    Max line width for wrapping markdown files.

    Can be a positive number, a string ``no``, or a string ``keep``. ``no`` will
    format all paragraphs in a single line, while ``keep`` will preserve the original
    wrapping.

    """

    change_categories: dict[str, str] = yuio.config.field(
        default={
            "security": "Security",
            "breaking": "Breaking",
            "added": "Added",
            "changed": "Changed",
            "deprecated": "Deprecated",
            "removed": "Removed",
            "performance": "Performance",
            "fixed": "Fixed",
        },
        help="ordering and titles for change categories",
    )
    """
    Ordering and titles for change categories.

    This config item defines known change categories (i.e. third level headings
    in release section, such as "Added" or "Changed").

    Its keys are category names used in other config items to refer to sections.

    Its values are category titles that will be used in section headings.
    If a value is empty, section's heading will not be replaced.

    Order of elements defines the preferred order of sections in the changelog.

    By default, changelog keeper defines the following sections:

    .. list-table::
        :header-rows: 1

        * - Category
          - Title
        * - ``security``
          - Security
        * - ``breaking``
          - Breaking
        * - ``added``
          - Added
        * - ``changed``
          - Changed
        * - ``deprecated``
          - Deprecated
        * - ``removed``
          - Removed
        * - ``performance``
          - Performance
        * - ``fixed``
          - Fixed

    :meta value: ...

    """

    extra_change_categories: dict[str, str] = yuio.app.field(
        default={},
        merge=lambda l, r: {**l, **r},
        help=(
            "additional items that will be added to `--cfg-change-categories` "
            "without completely overriding it"
        ),
    )
    """
    Additional items that will be added to :attr:`~Config.change_categories`
    without completely overriding it.

    """

    bump_patch_categories: set[str] = yuio.app.field(
        default={
            "security",
            "deprecated",
            "performance",
            "fixed",
        },
        help=(
            "if these change categories appear in unreleased section, "
            "suggest bumping the patch version component."
        ),
    )
    """
    If these change categories appear in unreleased section, suggest bumping
    the patch version component.

    Default is ``security``, ``deprecated``, ``performance``, ``fixed``.

    :meta value: ...

    """

    extra_bump_patch_categories: set[str] = yuio.app.field(
        default=set(),
        merge=lambda l, r: l | r,
        help=(
            "additional items that will be added to `--cfg-bump-patch-categories` "
            "without completely overriding it"
        ),
    )
    """
    Additional items that will be added to :attr:`~Config.bump_patch_categories`
    without completely overriding it.

    """

    bump_minor_categories: set[str] = yuio.app.field(
        default={
            "added",
            "changed",
            "removed",
        },
        help=(
            "if these change categories appear in unreleased section, "
            "suggest bumping the minor version component."
        ),
    )
    """
    If these change categories appear in unreleased section, suggest bumping
    the minor version component.

    Default is ``added``, ``changed``, ``removed``.

    :meta value: ...

    """

    extra_bump_minor_categories: set[str] = yuio.app.field(
        default=set(),
        merge=lambda l, r: l | r,
        help=(
            "additional items that will be added to `--cfg-bump-minor-categories` "
            "without completely overriding it"
        ),
    )
    """
    Additional items that will be added to :attr:`~Config.bump_minor_categories`
    without completely overriding it.

    """

    bump_major_categories: set[str] = yuio.app.field(
        default={
            "breaking",
        },
        help=(
            "if these change categories appear in unreleased section, "
            "suggest bumping the major version component."
        ),
    )
    """
    If these change categories appear in unreleased section, suggest bumping
    the major version component.

    Default is ``breaking``.

    :meta value: ...

    """

    extra_bump_major_categories: set[str] = yuio.app.field(
        default=set(),
        merge=lambda l, r: l | r,
        help=(
            "additional items that will be added to `--cfg-bump-major-categories` "
            "without completely overriding it"
        ),
    )
    """
    Additional items that will be added to :attr:`~Config.bump_major_categories`
    without completely overriding it.

    """

    change_categories_map: dict[str, str] = yuio.app.field(
        default={
            r"(?im)\bsecurity\b": "security",
            r"(?im)\bbreak(ing|s)?\b": "breaking",
            r"(?im)\badd(ed|s)?\b": "added",
            r"(?im)\bchange(d|s)?\b": "changed",
            r"(?im)\bdeprecate(d|s)?\b": "deprecated",
            r"(?im)\bremove(d|s)?\b": "removed",
            r"(?im)\bperformance\b": "performance",
            r"(?im)\bfix(ed|es)?\b": "fixed",
        },
        help=(
            "a mapping from regular expressions to change category names, "
            "used to parse and normalize categories"
        ),
    )
    r"""
    A mapping from regular expressions to change category names, used to parse
    and normalize categories.

    When processing a release, each third level heading will be matched against
    these regular expressions to figure out which section it represents. If there
    is a match, the heading content will be replaced by the one from
    :attr:`~Config.change_categories`, and the section will be reordered to match
    the order from :attr:`~Config.change_categories`.

    Regular expression syntax is described in the :mod:`re` module documentation.
    Regular expressions will be compiled without flags, but you can use inline
    flags to enable case-insensitivity.

    :meta value: ...

    """

    extra_change_categories_map: dict[str, str] = yuio.app.field(
        default={},
        merge=lambda l, r: {**l, **r},
        help=(
            "additional items that will be added to `--cfg-change-categories-map` "
            "without completely overriding it"
        ),
    )
    """
    Additional items that will be added to :attr:`~Config.change_categories_map`
    without completely overriding it.

    """

    unreleased_name: str = yuio.app.field(
        default="Unreleased",
        help="title for the `unreleased` section of the changelog",
    )
    """
    Title for the ``unreleased`` section of the changelog.

    """

    unreleased_decorations: tuple[str, str] = yuio.app.field(
        default=("", ""),
        help="symbols that will be added around unreleased heading",
    )
    """
    Symbols that will be added around :attr:`~Config.unreleased_name`, but won't
    be turned into a link.

    This value should be a tuple of two strings, first will be placed before the title,
    second will be placed after.

    """

    unreleased_pattern: str = yuio.app.field(
        default=r"(?i)unreleased",
        help="regular expression used to detect sections with unreleased changes",
    )
    """
    Regular expression used to detect sections with unreleased changes.

    """

    release_decorations: tuple[str, str] = yuio.app.field(
        default=("", ""),
        help="symbols that will be added around each release heading",
    )
    """
    Symbols that will be added around each release heading.

    """

    release_date_decorations: tuple[str, str] = yuio.app.field(
        default=(" - ", ""),
        help="symbols that will be added around release date",
    )
    """
    Symbols that will be added around release date.

    """

    version_decorations: tuple[str, str] = yuio.app.field(
        default=("", ""),
        help="symbols that will be added around release version",
    )
    """
    Symbols that will be added around release version.

    """

    release_comment_decorations: tuple[str, str] = yuio.app.field(
        default=(" - ", ""),
        help="symbols that will be added around release comments",
    )
    """
    Symbols that will be added around release comments.

    """

    add_release_date: bool = yuio.app.field(
        default=True,
        help="whether changelog entries should have a release date in their headings",
    )
    """
    Whether changelog entries should have a release date in their headings.

    """

    add_release_link: bool = yuio.app.field(
        default=True,
        help=(
            "whether changelog entries should have links to tags/diffs in their headings"
        ),
    )
    """
    Whether changelog entries should have links to tags/diffs in their headings.

    """

    release_link_preset: ReleaseLinkPreset | None = yuio.app.field(
        default=None,
        help="preset for release link templates",
    )
    """
    Preset for release link templates.

    By default, a correct preset is inferred by inspecting repository's remotes.
    If this process fails, you can configure a preset manually.

    """

    release_link_template: str | None = yuio.app.field(
        default=None,
        help=(
            "template for a release link, "
            "formatted using Python's `str.format` method"
        ),
    )
    """
    Template for a release link, formatted using Python's :meth:`str.format` method.
    This setting overrides link from :attr:`preset <Config.release_link_preset>`.

    The following template variables are available, in addition to
    ones declared in :attr:`~Config.release_link_template_vars`:

    ``tag``
        Git tag for the current release.

    ``prev_tag``
        Git tag for the previous release.

    **Example:**

    .. code-block:: yaml

        release_link_template: "https://{host}/{repo}/compare/{prev_tag}...{tag}"
        release_link_template_vars:
            host: "github.com"
            repo: "taminomara/cl-keeper"

    """

    release_link_template_last: str | None = yuio.app.field(
        default=None,
        help=(
            "template for a link for an unreleased entry, "
            "formatted using Python's `str.format` method"
        ),
    )
    """
    Template for a link for an unreleased entry, formatted using
    Python's :meth:`str.format` method. This setting overrides link from
    :attr:`preset <Config.release_link_preset>`.

    The following template variables are available, in addition to
    ones declared in :attr:`~Config.release_link_template_vars`:

    ``prev_tag``
        Git tag for the latest release.

    **Example:**

    .. code-block:: yaml

        release_link_template_last: "https://{host}/{repo}/compare/{prev_tag}...HEAD"
        release_link_template_vars:
            host: "github.com"
            repo: "taminomara/cl-keeper"

    """

    release_link_template_first: str | None = yuio.app.field(
        default=None,
        help=(
            "template for a link for the first release, "
            "formatted using Python's `str.format` method"
        ),
    )
    """
    Template for a link for the first release, formatted using
    Python's :meth:`str.format` method. This setting overrides link from
    :attr:`preset <Config.release_link_preset>`.

    The following template variables are available, in addition to
    ones declared in :attr:`~Config.release_link_template_vars`:

    ``tag``
        Git tag for the current release.

    **Example:**

    .. code-block:: yaml

        release_link_template_first: "https://{host}/{repo}/releases/tag/{tag}"
        release_link_template_vars:
            host: "github.com"
            repo: "taminomara/cl-keeper"

    """

    release_link_template_vars: dict[str, str] | None = yuio.app.field(
        default=None,
        help=("additional variables that will be available in release link templates"),
    )
    """
    Additional variables that will be available in release link templates.

    """

    check_repo_tags: bool = yuio.app.field(
        default=True,
        help="scan git repo and ensure that changelog doesn't miss any release",
    )
    """
    Scan git repo and ensure that changelog doesn't miss any release.

    Default is ``true``.

    """

    tag_prefix: str = yuio.app.field(
        default="v",
        help="prefix for release tags",
    )
    """
    Prefix for release tags.

    Only tags that start with this prefix are considered release tags.

    Default is ``"v"``.

    """

    version_format: VersionFormat = yuio.app.field(
        default=VersionFormat.SEMVER,
        help="versioning schema used to parse and sort tags and versions",
    )
    """
    Versioning schema used to parse and sort tags and versions.

    See :class:`VersionFormat` for details.

    """

    ignore_missing_releases_before: str | None = yuio.app.field(
        default=None,
        help="don't complain about missing releases for tags before this one",
    )
    """
    Don't complain about missing releases for tags before this one.

    """

    ignore_missing_releases_regexp: str | None = yuio.app.field(
        default=r"(?i)dev|b|beta|a|alpha|rc|post",
        help="don't complain about missing releases if version matches this regexp",
    )
    """
    Don't complain about missing releases if version matches this regexp.

    Default regexp searches for the following substrings:
    ``dev|b|beta|a|alpha|rc|post``.

    """

    severity: dict[
        _t.Annotated[
            IssueCode,
            yuio.parse.WithMeta(desc="<code>"),
        ],
        _t.Annotated[
            IssueSeverity,
            yuio.parse.Enum(by_name=True, to_dash_case=True),
            yuio.parse.WithMeta(desc="<severity>"),
        ],
    ] = yuio.config.field(
        default={},
        merge=lambda l, r: {**l, **r},
        help="mapping from issue code to issue severity",
    )
    """
    A mapping from issue code to issue severity. Allows customizing and disabling
    certain checks.

    """

    @functools.cached_property
    def full_change_categories(self):
        full_change_categories = self.change_categories.copy()
        full_change_categories.update(self.extra_change_categories)
        return full_change_categories

    @functools.cached_property
    def full_change_categories_map(self):
        full_change_categories_map = self.change_categories_map.copy()
        full_change_categories_map.update(self.extra_change_categories_map)
        return full_change_categories_map

    @functools.cached_property
    def change_categories_sort_keys(self):
        return {k: i for i, k in enumerate(self.full_change_categories)}

    @functools.cached_property
    def parsed_ignore_missing_releases_before(self):
        from cl_keeper.parse import parse_version

        if self.ignore_missing_releases_before is None:
            return None
        else:
            return parse_version(self.ignore_missing_releases_before, self)

    @functools.cached_property
    def full_bump_patch_categories(self):
        return self.bump_patch_categories | self.extra_bump_patch_categories

    @functools.cached_property
    def full_bump_minor_categories(self):
        return self.bump_minor_categories | self.extra_bump_minor_categories

    @functools.cached_property
    def full_bump_major_categories(self):
        return self.bump_major_categories | self.extra_bump_major_categories

    def validate_config(self):
        if self.release_link_template_vars:
            for var in ["prev_tag", "tag"]:
                if var in self.release_link_template_vars:
                    raise yuio.parse.ParsingError(
                        f"release_link_template_vars can't contain key {var!r}",
                    )
        for categories, name in [
            (self.full_change_categories_map.values(), "change_categories_map"),
            (self.full_bump_patch_categories, "bump_patch_categories"),
            (self.full_bump_minor_categories, "bump_minor_categories"),
            (self.full_bump_major_categories, "bump_major_categories"),
        ]:
            for category in categories:
                if category not in self.full_change_categories:
                    raise yuio.parse.ParsingError(
                        f"{name} has an entry for category {category}, "
                        f"which is missing from change_categories. Please add {category} "
                        "to change_categories",
                    )

        if (
            self.ignore_missing_releases_before is not None
            and self.parsed_ignore_missing_releases_before is None
        ):
            if self.version_format is VersionFormat.NONE:
                raise yuio.parse.ParsingError(
                    "ignore_missing_releases_before can't be used "
                    "with `version_format=none`",
                )
            else:
                raise yuio.parse.ParsingError(
                    "ignore_missing_releases_before doesn't follow "
                    f"{self.version_format.value} specification",
                )


PYTHON_PRESET = Config(version_format=VersionFormat.PYTHON)
"""
Default config that's used when `pyproject.toml` is detected.

"""



from __future__ import annotations

import re
import typing as _t

from cl_keeper.context import Context, IssueCode, IssueScope
from cl_keeper.fix import make_link_for_section
from cl_keeper.model import (
    Changelog,
    ReleaseSection,
    RepoVersion,
    Section,
    SubSectionType,
    UnreleasedSection,
)


def check(
    changelog: Changelog, ctx: Context, repo_versions: dict[str, RepoVersion] | None
):
    check_duplicates(changelog.sections, ctx)
    found_unordered_sections = check_order(changelog.sections, ctx)
    check_links(
        changelog.sections,
        repo_versions,
        ctx,
        not found_unordered_sections,
    )
    check_dates(changelog.sections, ctx, repo_versions)
    check_section_content(changelog.sections, ctx)
    if repo_versions is not None:
        check_tags(changelog.sections, ctx, repo_versions)


def check_duplicates(sections: _t.Iterable[Section], ctx: Context):
    seen_unreleased = False
    seen_versions: set[str] = set()

    for section in sections:
        if section.is_unreleased():
            if seen_unreleased:
                ctx.issue(
                    IssueCode.DUPLICATE_RELEASES,
                    "Found multiple sections for unreleased changes",
                    pos=section,
                )
            seen_unreleased = True
            check_duplicates_in_subsections(section, ctx)
        elif release := section.as_release():
            if release.canonized_version in seen_versions:
                ctx.issue(
                    IssueCode.DUPLICATE_RELEASES,
                    "Found multiple sections for release `%s`",
                    release.version,
                    pos=section,
                )
            seen_versions.add(release.canonized_version)
            check_duplicates_in_subsections(section, ctx)


def check_duplicates_in_subsections(section: Section, ctx: Context):
    seen_categories: set[str] = set()
    for subsection in section.subsections:
        match subsection.type:
            case SubSectionType.TRIVIA:
                pass
            case SubSectionType.CHANGES:
                if subsection.category in seen_categories:
                    ctx.issue(
                        IssueCode.DUPLICATE_CHANGE_CATEGORIES,
                        f"Found multiple sub-sections for `%s` category in {section.what()}",
                        subsection.category,
                        pos=subsection,
                    )
                else:
                    seen_categories.add(subsection.category)


def check_order(sections: list[Section], ctx: Context):
    if (
        not sections
        or not sections[0].is_trivia()
        or not sections[0].subsections
        or not sections[0].subsections[0].content
        or sections[0].subsections[0].content[0].tag != "h1"
    ):
        ctx.issue(
            IssueCode.GENERAL_FORMATTING_ERROR,
            "Changelog must start with a first level heading",
            pos=sections[0] if sections else None,
        )
        return
    last_version = None
    found_unordered_sections = False
    for section in sections:
        if section.is_unreleased():
            if last_version is not None:
                ctx.issue(
                    IssueCode.RELEASE_ORDERING,
                    "Section for unreleased changes must be first in the changelog",
                    pos=section,
                )
            check_order_in_subsection(section, ctx)
        elif release := section.as_release():
            if (
                last_version
                and release.parsed_version
                and last_version < release.parsed_version
            ):
                if not found_unordered_sections:
                    # Don't spam too many warnings.
                    ctx.issue(
                        IssueCode.RELEASE_ORDERING,
                        "Sections are not ordered by release versions",
                        pos=section,
                    )
                found_unordered_sections = True
            elif release.parsed_version is not None:
                last_version = release.parsed_version
            check_order_in_subsection(section, ctx)

    return found_unordered_sections


def check_order_in_subsection(section: Section, ctx: Context):
    last_sort_key = None
    for subsection in section.subsections:
        match subsection.type:
            case SubSectionType.TRIVIA:
                continue
            case SubSectionType.CHANGES:
                if (
                    last_sort_key is not None
                    and subsection.sort_key is not None
                    and last_sort_key > subsection.sort_key
                ):
                    ctx.issue(
                        IssueCode.CHANGE_CATEGORY_ORDERING,
                        f"Sub-sections in {section.what()} are not ordered by preferred order",
                        pos=subsection,
                    )
                    return
                elif subsection.sort_key is not None:
                    last_sort_key = subsection.sort_key


def check_links(
    sections: list[Section],
    repo_versions: dict[str, RepoVersion] | None,
    ctx: Context,
    can_trust_order: bool,
):
    if not ctx.config.add_release_link:
        for section in reversed(sections):
            if not isinstance(section, (ReleaseSection, UnreleasedSection)):
                continue
            if section.version_link is not None:
                ctx.issue(
                    IssueCode.UNEXPECTED_RELEASE_LINK,
                    f"Unexpected link for {section.what()}",
                    pos=section,
                )
        return

    prev_tag: str | None = None
    for section in reversed(sections):
        if not isinstance(section, (ReleaseSection, UnreleasedSection)):
            continue
        if section.is_unreleased() and not can_trust_order:
            continue

        canonical_link, prev_tag = make_link_for_section(
            section, prev_tag, repo_versions, ctx
        )

        if canonical_link is None:
            if section.version_link is not None:
                ctx.issue(
                    IssueCode.UNEXPECTED_RELEASE_LINK,
                    f"Unexpected link for {section.what()}",
                    pos=section,
                )
        else:
            if section.version_link is None:
                if can_trust_order:
                    ctx.issue(
                        IssueCode.MISSING_RELEASE_LINK,
                        f"Missing link for {section.what()}, should be <c path>%s</c>",
                        canonical_link,
                        pos=section,
                    )
                else:
                    ctx.issue(
                        IssueCode.MISSING_RELEASE_LINK,
                        f"Missing link for {section.what()}",
                        pos=section,
                    )
            elif section.version_link != canonical_link:
                if can_trust_order:
                    ctx.issue(
                        IssueCode.MISSING_RELEASE_LINK,
                        f"Incorrect link for {section.what()}, should be <c path>%s</c>",
                        canonical_link,
                        pos=section,
                    )
                else:
                    ctx.issue(
                        IssueCode.INCORRECT_RELEASE_LINK,
                        f"Potentially incorrect link for {section.what()}",
                        pos=section,
                    )


def check_dates(
    sections: list[Section],
    ctx: Context,
    repo_versions: dict[str, RepoVersion] | None,
):
    for section in sections:
        release = section.as_release()
        if not release:
            continue
        if ctx.config.add_release_date and release.release_date_fmt is None:
            ctx.issue(
                IssueCode.MISSING_RELEASE_DATE,
                "Missing date for release `%s`",
                release.version,
                pos=section,
            )
        elif (
            ctx.config.add_release_date
            and repo_versions is not None
            and release.release_date is not None
            and (data := repo_versions.get(release.canonized_version))
        ):
            if (
                data.author_date != release.release_date
                and data.committer_date != release.release_date
            ):
                ctx.issue(
                    IssueCode.INCORRECT_RELEASE_DATE,
                    "Release date for release `%s` is different from commit date `%s`",
                    release.version,
                    data.committer_date.isoformat(),
                    pos=release,
                )
        elif not ctx.config.add_release_date and release.release_date_fmt is None:
            ctx.issue(
                IssueCode.UNEXPECTED_RELEASE_DATE,
                "Unexpected release date for release `%s`",
                release.version,
                pos=release,
            )


def check_section_content(sections: list[Section], ctx: Context):
    is_first_section = True
    lower_bound = ctx.config.parsed_ignore_missing_releases_before
    regex_bound = ctx.config.ignore_missing_releases_regexp
    for section in sections:
        nodes = section.walk()
        if section.is_trivia() and is_first_section:
            try:
                next(nodes)  # Skip first <h1>.
            except StopIteration:
                pass
        is_first_section = False
        for node in nodes:
            if node.tag == "h1":
                ctx.issue(
                    IssueCode.GENERAL_FORMATTING_ERROR,
                    "Unexpected first level heading",
                    pos=node,
                )
        if release := section.as_release():
            if (
                not release.subsections
                and release.parsed_version is not None
                and (lower_bound is None or release.parsed_version >= lower_bound)
                and (regex_bound is None or not re.search(regex_bound, release.version))
            ):
                ctx.issue(
                    IssueCode.EMPTY_RELEASE,
                    "Section for release `%s` is empty",
                    release.version,
                    pos=section,
                )
            for subsection in release.subsections:
                if not subsection.content:
                    ctx.issue(
                        IssueCode.EMPTY_CHANGE_CATEGORY,
                        "Sub-section `%s` for release `%s` is empty",
                        subsection.category,
                        release.version,
                        pos=section,
                    )
                    break


def check_tags(
    sections: list[Section], ctx: Context, repo_versions: dict[str, RepoVersion]
):
    known_versions = set(repo_versions)
    releases = set(
        release.canonized_version
        for section in sections
        if (release := section.as_release())
    )
    if not_in_repo := releases - known_versions:
        ctx.issue(
            IssueCode.MISSING_TAG_FOR_RELEASE,
            f"Missing tags for release{"" if len(not_in_repo) == 1 else "s"} "
            f"{_join_more(not_in_repo)}",
            scope=IssueScope.EXTERNAL,
        )
    lower_bound = ctx.config.parsed_ignore_missing_releases_before
    regex_bound = ctx.config.ignore_missing_releases_regexp
    not_in_changelog = []
    for data in repo_versions.values():
        if (
            data.canonized_version not in releases
            and data.parsed_version is not None
            and (lower_bound is None or data.parsed_version >= lower_bound)
            and (
                regex_bound is None
                or not re.search(regex_bound, data.canonized_version)
            )
        ):
            not_in_changelog.append(data.version)
    if not_in_changelog:
        ctx.issue(
            IssueCode.MISSING_RELEASE_FOR_TAG,
            f"Missing changelog sections for "
            f"release{"" if len(not_in_changelog) == 1 else "s"} "
            f"{_join_more(not_in_changelog)}",
            scope=IssueScope.EXTERNAL,
        )


def _join_more(strings: _t.Collection[str]) -> str:
    joined = ", ".join(f"`{s}`" for s in sorted(strings)[:5])
    if len(strings) > 5:
        joined += f" (+{len(strings) - 5} more)"
    return joined



import datetime
import re
import typing as _t

import mdformat.plugins
import mdformat_footnote  # type: ignore
import mdformat_frontmatter  # type: ignore
import mdformat_gfm._mdformat_plugin  # type: ignore
import mdformat_gfm_alerts
import mdformat_tables  # type: ignore
import packaging.version
import semver
from markdown_it import MarkdownIt
from markdown_it.tree import SyntaxTreeNode
from mdformat.renderer import MDRenderer

from cl_keeper.config import Config, VersionFormat
from cl_keeper.context import Context, IssueCode
from cl_keeper.fix import format_section_heading_text
from cl_keeper.model import (
    Changelog,
    ReleaseSection,
    Section,
    SubSection,
    SubSectionCategoryKind,
    SubSectionType,
    UnreleasedSection,
    Version,
)

_VERSION_RE = re.compile(
    r"""
        (?:
            (?:\d+!)?
            \d+\.\d+\.\d+
            (?:[+-._]?[0-9a-zA-Z-+.]*)?
        )
    """,
    re.VERBOSE | re.IGNORECASE,
)
_DATE_RE = re.compile(
    r"""
        (?P<year>\d\d\d\d)
        -
        (?P<month>\d\d)
        -
        (?P<day>\d\d)
    """,
    re.VERBOSE,
)


def parse(ctx: Context) -> Changelog:
    """
    Parse a changelog.

    """

    parser = build_parser()
    parser_env = {"references": {}}

    tokens = parser.parse(ctx.src, parser_env)
    root = SyntaxTreeNode(tokens)

    sections = build_changelog(root, ctx)

    return Changelog(
        sections=sections,
        parser=parser,
        parser_env=parser_env,
    )


def build_parser() -> MarkdownIt:
    parser = MarkdownIt(renderer_cls=MDRenderer)  # pyright: ignore[reportArgumentType]
    parser.options["store_labels"] = True
    parser.options["parser_extension"] = []
    parser.options["mdformat"] = {}

    # Note: entrypoint discovery system doesn't work in standalone nuitka builds.
    # We load all plugins manually.
    plugins = [
        ("footnote", mdformat_footnote),
        ("frontmatter", mdformat_frontmatter),
        ("tables", mdformat_tables),
        ("gfm_alerts", mdformat_gfm_alerts),
        ("gfm", mdformat_gfm._mdformat_plugin),
    ]

    for name, plugin in plugins:
        mdformat.plugins.PARSER_EXTENSIONS.setdefault(name, plugin)  # type: ignore
        if plugin not in parser.options["parser_extension"]:
            parser.options["parser_extension"].append(plugin)
            plugin.update_mdit(parser)

    return parser


def build_changelog(root: SyntaxTreeNode, ctx: Context) -> list[Section]:
    """
    Build changelog sections from content.

    """

    sections: list[Section] = []
    for heading, content in split_into_sections(root.children, 2):
        subsections: list[SubSection] = []
        if not heading:
            subsections.append(SubSection(content=content))
        else:
            for subheading, content in split_into_sections(content, 3):
                subsection = SubSection(heading=subheading, content=content)
                detect_subsection_metadata(subsection, ctx)
                subsections.append(subsection)
        sections.append(create_section(heading, subsections, ctx))
    return sections


def split_into_sections(nodes: _t.Iterable[SyntaxTreeNode], target_heading_level: int):
    """
    Group nodes into sections, splitting them at heading of the given level.

    """

    sections: list[tuple[SyntaxTreeNode | None, list[SyntaxTreeNode]]] = []

    current_heading: SyntaxTreeNode | None = None
    current_section: list[SyntaxTreeNode] = []

    for node in nodes:
        if node.type != "heading":
            current_section.append(node)
            continue
        level = int(node.tag[1:])
        if level > target_heading_level:
            current_section.append(node)
            continue
        else:
            if current_heading or current_section:
                sections.append((current_heading, current_section))
            if level == target_heading_level:
                current_heading = node
                current_section = []
            else:
                current_heading = None
                current_section = [node]
    if current_heading or current_section:
        sections.append((current_heading, current_section))
    return sections


def create_section(
    heading: SyntaxTreeNode | None, subsections: list[SubSection], ctx: Context
) -> Section:
    """
    Scan section heading and set its type, version, release date, and so on.

    """

    heading_text = _node_to_text(heading)

    if heading is None or heading.tag != "h2" or heading_text is None:
        return Section(
            heading=heading,
            subsections=subsections,
        )

    link: str | None = None
    label: str | None = None
    for node in heading.walk():
        if node.type == "link":
            link = str(node.attrs.get("href", "")) or None
            label = node.meta.get("label")
            break

    if re.search(ctx.config.unreleased_pattern, heading_text):
        section = UnreleasedSection(
            heading=heading,
            subsections=subsections,
            version_link=link,
            version_label=label,
        )

        canonical_heading = format_section_heading_text(section, ctx)
        if heading_text != canonical_heading:
            ctx.issue(
                IssueCode.UNRELEASED_HEADING_FORMAT,
                "Heading for unreleased changes isn't properly formatted, should be `%s`",
                canonical_heading,
                pos=section.heading,
            )

        return section

    version_match = _VERSION_RE.search(heading_text)
    if not version_match:
        section = Section(
            heading=heading,
            subsections=subsections,
        )

        ctx.issue(
            IssueCode.GENERAL_FORMATTING_ERROR,
            "Changelog section doesn't contain a release version",
            pos=section.heading,
        )

        return section

    version = version_match.group(0)

    parsed_version = parse_version(version, ctx.config)
    if parsed_version is None and ctx.config.version_format is not VersionFormat.NONE:
        ctx.issue(
            IssueCode.INVALID_VERSION,
            "Version `%s` doesn't follow %s specification",
            version,
            ctx.config.version_format.value,
            pos=heading,
        )
    canonized_version = canonize_version(parsed_version, ctx.config) or version

    date_match = _DATE_RE.search(heading_text)
    if date_match is None:
        release_date = None
        release_date_fmt = None
    else:
        release_date_fmt = date_match.group()
        try:
            release_date = datetime.date(
                year=int(date_match.group("year")),
                month=int(date_match.group("month")),
                day=int(date_match.group("day")),
            )
        except ValueError as e:
            release_date = None
            ctx.issue(
                IssueCode.INVALID_RELEASE_DATE,
                "Incorrect release date `%s`: %s",
                release_date_fmt,
                e,
                pos=heading,
            )

    found_unresolved_link = False
    if date_match is None:
        prefix = heading_text[: version_match.start()]
        suffix = heading_text[version_match.end() :]

        prefix = prefix.removesuffix(ctx.config.version_decorations[0])
        suffix = suffix.removeprefix(ctx.config.version_decorations[1])

        if check_broken_link(prefix, suffix):
            ctx.issue(
                IssueCode.RELEASE_HEADING_FORMAT,
                "Can't resolve version's link",
                pos=heading,
            )
            suffix = suffix.lstrip().removeprefix("]")
            found_unresolved_link = True
    elif version_match.end() <= date_match.start():
        prefix = heading_text[: version_match.start()]
        middle = heading_text[version_match.end() : date_match.start()]
        suffix = heading_text[date_match.end() :]

        prefix = prefix.removesuffix(ctx.config.version_decorations[0])
        middle = middle.removeprefix(ctx.config.version_decorations[1])
        suffix = suffix.removeprefix(ctx.config.release_date_decorations[1])

        if check_broken_link(prefix, middle):
            ctx.issue(
                IssueCode.RELEASE_HEADING_FORMAT,
                "Can't resolve version's link",
                pos=heading,
            )
            found_unresolved_link = True

        if middle.endswith(("(", "[", "<", "{")) and suffix.startswith(
            (")", "]", ">", "}")
        ):
            suffix = suffix[1:]
    else:
        prefix = heading_text[: date_match.start()]
        middle = heading_text[date_match.end() : version_match.start()]
        suffix = heading_text[version_match.end() :]

        middle = middle.removesuffix(ctx.config.version_decorations[0])
        suffix = suffix.removeprefix(ctx.config.version_decorations[1])

        if check_broken_link(middle, suffix):
            ctx.issue(
                IssueCode.RELEASE_HEADING_FORMAT,
                "Can't resolve version's link",
                pos=heading,
            )
            suffix = suffix.lstrip().removeprefix("]")
            found_unresolved_link = True

        if middle.endswith(("(", "[", "<", "{")) and suffix.startswith(
            (")", "]", ">", "}")
        ):
            suffix = suffix[1:]

    release_comment = (
        suffix.removeprefix(ctx.config.release_comment_decorations[0])
        .removesuffix(ctx.config.release_comment_decorations[1])
        .strip()
    )

    section = ReleaseSection(
        heading=heading,
        subsections=subsections,
        version=version,
        parsed_version=parsed_version,
        canonized_version=canonized_version,
        version_link=link,
        version_label=label,
        release_date=release_date,
        release_date_fmt=release_date_fmt,
        release_comment=release_comment,
    )

    if not found_unresolved_link:
        canonical_heading = format_section_heading_text(section, ctx)
        if canonical_heading != heading_text:
            ctx.issue(
                IssueCode.RELEASE_HEADING_FORMAT,
                "Heading for release `%s` isn't properly formatted, should be `%s`",
                section.version,
                canonical_heading,
                pos=section.heading,
            )

    return section


def check_broken_link(prefix: str, suffix: str):
    return prefix.rstrip().endswith("[") and suffix.lstrip().startswith("]")


def detect_subsection_metadata(subsection: SubSection, ctx: Context):
    """
    Scan subsection heading and set its type, category, and sort key.

    """

    heading = _node_to_text(subsection.heading)

    if subsection.heading is None or subsection.heading.tag != "h3" or heading is None:
        subsection.type = SubSectionType.TRIVIA
        subsection.category_kind = SubSectionCategoryKind.UNKNOWN
        subsection.category = ""
        subsection.sort_key = None
        return

    for regex, category in ctx.config.full_change_categories_map.items():
        if re.search(regex, heading):
            subsection.type = SubSectionType.CHANGES
            subsection.category_kind = SubSectionCategoryKind.KNOWN
            subsection.category = category
            subsection.sort_key = ctx.config.change_categories_sort_keys.get(category)

            canonical_heading = ctx.config.full_change_categories.get(category)
            if canonical_heading and canonical_heading != heading:
                ctx.issue(
                    IssueCode.CHANGE_CATEGORY_HEADING_FORMAT,
                    "Heading for change group isn't properly formatted, should be `%s`",
                    canonical_heading,
                    pos=subsection.heading,
                )

            return

    # A group with h3-level heading is always a change category,
    # albeit not a recognized one.
    subsection.type = SubSectionType.CHANGES
    subsection.category_kind = SubSectionCategoryKind.UNKNOWN
    subsection.category = heading.casefold()
    subsection.sort_key = None

    ctx.issue(
        IssueCode.UNKNOWN_CHANGE_CATEGORY,
        "Unknown change group `%s`",
        heading,
        pos=subsection.heading,
    )


def _node_to_text(heading: SyntaxTreeNode | None):
    if heading is None:
        return None
    text = ""
    for node in heading.walk(include_self=False):
        if node.type == "text":
            text += node.content
    return text.strip()


_STRICT_SEMVER_TEMPLATE = re.compile(
    r"""
        ^
        (?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)
        (?:-(?:alpha|beta|rc)(?:0|[1-9]\d*))?
        $
    """,
    re.VERBOSE,
)

_PYTHON_SEMVER_TEMPLATE = re.compile(
    r"""
        ^
        (?:(?:0|[1-9]\d*)!)?
        (?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)
        (?:
            -(?:
                (?:alpha|beta|rc)(?:0|[1-9]\d*)(?:\.post(?:0|[1-9]\d*))?
                | (?:post(?:0|[1-9]\d*))
            )
        )?
        $
    """,
    re.VERBOSE,
)

_STRICT_PYTHON_TEMPLATE = re.compile(
    r"""
        ^
        (?:(?:0|[1-9]\d*)!)?
        (?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)
        (?:(?:a|b|rc)(?:0|[1-9]\d*))?
        (?:\.post(?:0|[1-9]\d*))?
        $
    """,
    re.VERBOSE,
)


def parse_version(version: str | None, config: Config) -> Version | None:
    if version is None:
        return None
    try:
        match config.version_format:
            case VersionFormat.SEMVER:
                return _t.cast(Version, semver.Version.parse(version))
            case VersionFormat.SEMVER_STRICT:
                if not _STRICT_SEMVER_TEMPLATE.match(version):
                    return None
                return _t.cast(Version, semver.Version.parse(version))
            case VersionFormat.PYTHON:
                return _t.cast(Version, packaging.version.Version(version))
            case VersionFormat.PYTHON_STRICT:
                if not _STRICT_PYTHON_TEMPLATE.match(version):
                    return None
                return _t.cast(Version, packaging.version.Version(version))
            case VersionFormat.PYTHON_SEMVER:
                if not _PYTHON_SEMVER_TEMPLATE.match(version):
                    return None
                return _t.cast(Version, packaging.version.Version(version))
            case VersionFormat.NONE:
                return None
    except ValueError:
        return None


def canonize_version(version: Version | None, config: Config) -> str | None:
    if version is None:
        return None
    match config.version_format:
        case VersionFormat.SEMVER:
            assert isinstance(version, semver.Version)
            return str(version)
        case VersionFormat.SEMVER_STRICT:
            assert isinstance(version, semver.Version)
            return str(version)
        case VersionFormat.PYTHON:
            assert isinstance(version, packaging.version.Version)
            return str(version)
        case VersionFormat.PYTHON_STRICT:
            assert isinstance(version, packaging.version.Version)
            return str(version)
        case VersionFormat.PYTHON_SEMVER:
            assert isinstance(version, packaging.version.Version)

            parts = []

            # Epoch
            if version.epoch != 0:
                parts.append(f"{version.epoch}!")

            # Release segment
            assert len(version.release) <= 3
            parts.append(
                ".".join(
                    str(version.release[i] if len(version.release) >= i else 0)
                    for i in range(3)
                )
            )

            # Pre-release
            if version.pre is not None:
                match version.pre[0]:
                    case "a":
                        parts.append("-alpha")
                        parts.append(str(version.pre[1]))
                    case "b":
                        parts.append("-beta")
                        parts.append(str(version.pre[1]))
                    case "rc":
                        parts.append("-rc")
                        parts.append(str(version.pre[1]))
                    case _:
                        assert False

            # Post-release
            if version.post is not None:
                if version.pre is not None:
                    parts.append(".post")
                else:
                    parts.append("-post")
                parts.append(str(version.post))

            # Development release
            assert version.dev is None

            # Local version segment
            assert version.local is None

            return "".join(parts)
        case VersionFormat.NONE:
            return None



import datetime
import json

import sphinx.application
import sphinx.builders
import sphinx.builders.html
import yuio.json_schema

import cl_keeper
import cl_keeper.config

# -- Project information -----------------------------------------------------

project = "Changelog Keeper CLI"
copyright = f"{datetime.date.today().year}, Tamika Nomara"
author = "Tamika Nomara"
release = version = cl_keeper.__version__


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx_design",
    "yuio.ext.sphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}
# nitpick_ignore_regex = [
#     (r"py:class", r"(.*\.)?([A-Z]{1,2}|[A-Z]+_co|Cmp|SupportsLt|Sz|TAst|_[^.]*)")
# ]
autodoc_typehints_format = "short"
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = False

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["extra.css"]
html_theme_options = {
    "source_repository": "https://github.com/taminomara/cl-keeper",
    "source_branch": "main",
    "source_directory": "docs/source",
}


def on_write_started(app: sphinx.application.Sphinx, builder):
    if not isinstance(builder, sphinx.builders.html.StandaloneHTMLBuilder):
        return

    ctx = yuio.json_schema.JsonSchemaContext()
    schema = yuio.json_schema.Meta(
        cl_keeper.config.Config.to_json_schema(ctx),
        title=cl_keeper.config.Config.__name__,
        description=cl_keeper.config.Config.__doc__,
    )
    app.outdir.joinpath("schema.json").write_text(
        json.dumps(
            ctx.render(
                schema, id="https://cl-keeper.readthedocs.io/en/latest/schema.json"
            ),
            indent=2,
        )
    )


def setup(app: sphinx.application.Sphinx):
    app.connect("write-started", on_write_started)

    return {
        "version": cl_keeper.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import logging

import pytest
import yuio.exec

from cl_keeper.config import ReleaseLinkPreset
from cl_keeper.vcs import detect_origin


@pytest.mark.parametrize(
    "origin,expected",
    [
        (None, None),
        ("https://example.com/", None),
        (
            "git+https://github.com/example/test.git",
            ReleaseLinkPreset.GITHUB.get_links().update_vars(
                dict(host="github.com", repo="example/test"), override=True
            ),
        ),
        (
            "git+https://gist.github.com/example/test.git",
            ReleaseLinkPreset.GITHUB.get_links().update_vars(
                dict(host="gist.github.com", repo="example/test"), override=True
            ),
        ),
        (
            "git+https://gitlab.com/example/test.git",
            ReleaseLinkPreset.GITLAB.get_links().update_vars(
                dict(host="gitlab.com", repo="example/test"), override=True
            ),
        ),
        pytest.param(
            "git+https://bitbucket.org/example/test.git",
            None,
            marks=pytest.mark.xfail(
                reason="https://github.com/nephila/giturlparse/pull/108"
            ),
        ),
        ("git+https://friendco.de/owner@user/repo.git", None),
    ],
)
def test_detect_origin(origin, expected, tmp_path):
    dir = tmp_path / "repo"
    dir.mkdir()
    yuio.exec.exec("git", "init", str(dir), cwd=dir)
    if origin:
        yuio.exec.exec(
            "git", "remote", "add", "remote", origin, level=logging.INFO, cwd=dir
        )
    result = detect_origin(dir)
    assert result == expected



from __future__ import annotations

import math
import typing as _t

from cl_keeper.model import Section, SubSection
from cl_keeper.typing import Ord, SupportsPos


def sorted_sections(sections: _t.Iterable[Section]) -> list[Section]:
    """
    Sort sections by their release version.

    """

    return sorted_by_key(sections, _sections_key, reverse=True)


def sorted_subsections(subsections: _t.Iterable[SubSection]):
    """
    Sort subsections by their preferred order.

    """

    return sorted_by_key(subsections, lambda s: s.sort_key)


def _sections_key(section: Section):
    if section.is_unreleased():
        return (1, None)
    elif release := section.as_release():
        return (0, release.parsed_version) if release.parsed_version else None
    else:
        return None


def sorted_by_key[T: SupportsPos, K: Ord](
    items: _t.Iterable[T], key: _t.Callable[[T], K | None], reverse: bool = False
) -> list[T]:
    """
    Sort sections or subsections by the given key. If key is not present in an item,
    try keeping its position as close to the original as possible.

    """

    orderable: list[tuple[K, T]] = []
    unorderable: list[T] = []

    for item in items:
        if (k := key(item)) is not None:
            orderable.append((k, item))
        else:
            unorderable.append(item)

    orderable.sort(key=lambda x: x[0], reverse=reverse)

    if not unorderable:
        return [v for _, v in orderable]
    elif not orderable:
        return unorderable

    # Merge ordered and unordered items based on their position in the source.
    result: list[T] = []
    i, j = 0, 0
    while i < len(orderable) or j < len(unorderable):
        left = right = None
        if i < len(orderable):
            left = orderable[i][1]
        if j < len(unorderable):
            right = unorderable[j]
        if left and right:
            left_pos = left.map or (math.inf, 0)
            right_pos = right.map or (math.inf, 0)
            if left_pos <= right_pos:
                result.append(left)
                i += 1
            else:
                result.append(right)
                j += 1
        elif left:
            result.append(left)
            i += 1
        elif right:
            result.append(right)
            j += 1

    return result


def merge_sections(lhs: Section, rhs: Section):
    """
    Merge content of the right section into the left section.

    """

    subsections: dict[str | None, SubSection] = {}
    for section in lhs, rhs:
        for subsection in section.subsections:
            if subsection.category in subsections:
                merge_subsections(subsections[subsection.category], subsection)
            else:
                subsections[subsection.category] = subsection
    lhs.subsections = list(subsections.values())


def merge_subsections(lhs: SubSection, rhs: SubSection):
    """
    Merge content of the right subsection into the left subsection.

    """

    if (
        len(lhs.content) == 1
        and len(rhs.content) == 1
        and lhs.content[0].type in ["bullet_list", "ordered_list"]
        and rhs.content[0].type == lhs.content[0].type
    ):
        lhs.content[0].children += rhs.content[0].children
    else:
        lhs.content += rhs.content



import logging
import pathlib

import giturlparse
import yuio.git

from cl_keeper.config import (
    IssueCode,
    LinkTemplates,
    ReleaseLinkPreset,
    VersionFormat,
)
from cl_keeper.context import Context, IssueScope
from cl_keeper.model import RepoVersion
from cl_keeper.parse import canonize_version, parse_version

logger = logging.getLogger(__name__)


def detect_origin(root: pathlib.Path) -> LinkTemplates | None:
    try:
        url = yuio.git.Repo(root).git("ls-remote", "--get-url").decode().strip()
    except (yuio.git.GitError, UnicodeDecodeError) as e:
        logger.debug("failed to get origin url from git: %s", e)
        return None
    parsed = giturlparse.parse(url)
    if not parsed.valid:
        logger.debug("failed to parse origin url from git: %r", url)
        return None
    vars = {
        "repo": parsed.pathname.removesuffix(".git").strip("/"),
        "host": parsed.host,
    }
    logger.debug("detected origin url's platform as %r: %s", parsed.platform, url)
    match parsed.platform:
        case "github":
            logger.info(
                "using link preset 'github' with host %r, repo %r",
                vars["host"],
                vars["repo"],
            )
            return ReleaseLinkPreset.GITHUB.get_links().update_vars(vars, override=True)
        case "gitlab":
            logger.info(
                "using link preset 'gitlab' with host %r, repo %r",
                vars["host"],
                vars["repo"],
            )
            return ReleaseLinkPreset.GITLAB.get_links().update_vars(vars, override=True)
        case _:
            logger.info(
                "failed to parse origin url from git: platform %r is not supported",
                parsed.platform,
            )
            return None


def get_repo_versions(root: pathlib.Path, ctx: Context) -> dict[str, RepoVersion]:
    repo = yuio.git.Repo(root)
    repo_versions: dict[str, RepoVersion] = {}
    for commit in repo.log("--tags", "--no-walk"):
        for tag in commit.tags:
            if tag.startswith(ctx.config.tag_prefix):
                version = tag[len(ctx.config.tag_prefix) :]
                parsed_version = parse_version(version, ctx.config)
                if (
                    parsed_version is None
                    and ctx.config.version_format is not VersionFormat.NONE
                ):
                    ctx.issue(
                        IssueCode.INVALID_TAG,
                        "Tag `%s` doesn't follow %s specification",
                        tag,
                        ctx.config.version_format.value,
                        scope=IssueScope.EXTERNAL,
                    )
                canonized_version = (
                    canonize_version(parsed_version, ctx.config) or version
                )
                repo_versions[canonized_version] = RepoVersion(
                    version=version,
                    parsed_version=parsed_version,
                    canonized_version=canonized_version,
                    author_date=commit.author_datetime.date(),
                    committer_date=commit.committer_datetime.date(),
                )
    return repo_versions



from markdown_it.token import Token
from markdown_it.tree import SyntaxTreeNode

from cl_keeper.context import Context
from cl_keeper.model import (
    Changelog,
    ReleaseSection,
    RepoVersion,
    Section,
    SubSection,
    SubSectionCategoryKind,
    SubSectionType,
    UnreleasedSection,
)
from cl_keeper.sort import sorted_sections, sorted_subsections


def fix(
    changelog: Changelog, ctx: Context, repo_versions: dict[str, RepoVersion] | None
):
    changelog.sections = sorted_sections(changelog.sections)

    if (
        not changelog.sections
        or not changelog.sections[0].is_trivia()
        or not changelog.sections[0].subsections
        or not changelog.sections[0].subsections[0].content
        or changelog.sections[0].subsections[0].content[0].tag != "h1"
    ):
        changelog.sections.insert(
            0,
            Section(
                heading=None,
                subsections=[
                    SubSection(
                        type=SubSectionType.TRIVIA,
                        content=[format_heading("Changelog")],
                    )
                ],
            ),
        )

    prev_tag: str | None = None
    for section in reversed(changelog.sections):
        if not isinstance(section, (ReleaseSection, UnreleasedSection)):
            continue

        if (
            ctx.config.add_release_date
            and (release := section.as_release())
            and not release.release_date
            and repo_versions is not None
            and (data := repo_versions.get(release.canonized_version))
        ):
            release.release_date = data.author_date

        if ctx.config.add_release_link:
            section.version_link, prev_tag = make_link_for_section(
                section, prev_tag, repo_versions, ctx
            )
            section.version_label = format_section_label(section, ctx)
            if section.version_label:
                changelog.parser_env["references"][section.version_label] = {
                    "title": "",
                    "href": section.version_link,
                    "map": None,
                }

        section.heading = format_section_heading(section, ctx)

        section.subsections = sorted_subsections(section.subsections)
        for subsection in reversed(section.subsections):
            if subsection.category_kind is SubSectionCategoryKind.KNOWN and (
                heading := ctx.config.full_change_categories[subsection.category]
            ):
                subsection.heading = format_heading(heading, 3)


def format_heading(text: str, level: int = 1):
    return SyntaxTreeNode(
        [
            Token("heading_open", f"h{level}", 1, block=True, markup="#" * level),
            Token(
                "inline",
                "",
                0,
                children=[Token("text", "", 0, content=text)],
            ),
            Token("heading_close", f"h{level}", -1, block=True),
        ],
        create_root=False,
    )


def format_section_heading_text(section: Section, ctx: Context) -> str:
    version_pre, version_post = ctx.config.version_decorations
    release_date_pre, release_date_post = ctx.config.release_date_decorations
    release_comment_pre, release_comment_post = ctx.config.release_comment_decorations

    heading = ""

    if section.is_unreleased():
        heading += ctx.config.unreleased_decorations[0]
        heading += ctx.config.unreleased_name
        heading += ctx.config.unreleased_decorations[1]
    elif release := section.as_release():
        heading += ctx.config.release_decorations[0]
        heading += f"{version_pre}{release.version}{version_post}"
        if ctx.config.add_release_date:
            if release.release_date:
                release_date = release.release_date.isoformat()
            else:
                release_date = release.release_date_fmt
            if release_date is not None:
                heading += f"{release_date_pre}{release_date}{release_date_post}"
        if release.release_comment:
            heading += (
                f"{release_comment_pre}{release.release_comment}{release_comment_post}"
            )
        heading += ctx.config.release_decorations[1]

    return heading


def format_section_heading(section: Section, ctx: Context) -> SyntaxTreeNode:
    version_pre, version_post = ctx.config.version_decorations
    release_date_pre, release_date_post = ctx.config.release_date_decorations
    release_comment_pre, release_comment_post = ctx.config.release_comment_decorations

    inline: list[Token] = []

    if unreleased := section.as_unreleased():
        if ctx.config.unreleased_decorations[0]:
            inline.append(
                Token("text", "", 0, content=ctx.config.unreleased_decorations[0])
            )
        if unreleased.version_link and ctx.config.add_release_link:
            inline.append(
                Token(
                    "link_open",
                    "a",
                    1,
                    attrs={"href": unreleased.version_link},
                    meta={"label": unreleased.version_label},
                )
            )
        inline.append(Token("text", "", 0, content=ctx.config.unreleased_name))
        if unreleased.version_link and ctx.config.add_release_link:
            inline.append(Token("link_close", "a", -1))
        if ctx.config.unreleased_decorations[1]:
            inline.append(
                Token("text", "", 0, content=ctx.config.unreleased_decorations[1])
            )
    elif release := section.as_release():
        if ctx.config.release_decorations[0]:
            inline.append(
                Token("text", "", 0, content=ctx.config.release_decorations[0])
            )

        if release.version_link and ctx.config.add_release_link:
            inline.append(
                Token(
                    "link_open",
                    "a",
                    1,
                    attrs={"href": release.version_link},
                    meta={"label": release.version_label},
                )
            )
        inline.append(
            Token(
                "text", "", 0, content=f"{version_pre}{release.version}{version_post}"
            )
        )
        if release.version_link and ctx.config.add_release_link:
            inline.append(Token("link_close", "a", -1))

        if ctx.config.add_release_date:
            if release.release_date:
                release_date = release.release_date.isoformat()
            else:
                release_date = release.release_date_fmt
            if release_date is not None:
                inline.append(
                    Token(
                        "text",
                        "",
                        0,
                        content=f"{release_date_pre}{release_date}{release_date_post}",
                    )
                )
        if release.release_comment:
            inline.append(
                Token(
                    "text",
                    "",
                    0,
                    content=f"{release_comment_pre}{release.release_comment}{release_comment_post}",
                )
            )

        if ctx.config.release_decorations[1]:
            inline.append(
                Token("text", "", 0, content=ctx.config.release_decorations[1])
            )

    return SyntaxTreeNode(
        [
            Token("heading_open", "h2", 1, block=True, markup="##"),
            Token("inline", "", 0, children=inline),
            Token("heading_close", "h2", -1, block=True),
        ],
        create_root=False,
    )


def format_section_label(section: Section, ctx: Context):
    if section.is_unreleased():
        return ctx.config.unreleased_name
    elif release := section.as_release():
        pre, post = ctx.config.version_decorations
        return f"{pre}{release.version}{post}"
    else:
        return None


def make_link_for_section(
    section: Section,
    prev_tag: str | None,
    repo_versions: dict[str, RepoVersion] | None,
    ctx: Context,
):
    if section.is_unreleased():
        if prev_tag is not None:
            return (
                ctx.link_templates.template_last.format(
                    **ctx.link_templates.vars,
                    prev_tag=prev_tag,
                ),
                prev_tag,
            )
        else:
            return None, None
    elif release := section.as_release():
        if repo_versions is not None and (
            data := repo_versions.get(release.canonized_version)
        ):
            tag = f"{ctx.config.tag_prefix}{data.version}"
        else:
            tag = f"{ctx.config.tag_prefix}{release.version}"
        if prev_tag:
            return (
                ctx.link_templates.template.format(
                    **ctx.link_templates.vars,
                    tag=tag,
                    prev_tag=prev_tag,
                ),
                tag,
            )
        else:
            return (
                ctx.link_templates.template_first.format(
                    **ctx.link_templates.vars, tag=tag
                ),
                tag,
            )
    else:
        return None, None
